/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: SOAPMessage.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#if defined(WIN32)
#include <io.h>
#include <winsock.h>
#else
#include <unistd.h>
#include <netdb.h>
#endif

#include <iostream>
#include <algorithm>
#include <string>
#include <vector>

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif /* HAVE_CONFIG_H */

/* libxml2 */
#if defined(HAVE_LIBXML2_XMLVERSION_H) 
#include <libxml2/tree.h>
#include <libxml2/parser.h>
#else
#include <libxml/tree.h>
#include <libxml/parser.h>
#endif /* !HAVE_LIBXML2_XMLVERSION_H */

#include "SOAPMessage.h"
#include "MsgAttrHandler.h"
#include "SOAPException.h"
#include "XmlModifier.h"
#include "StringUtil.h"

#include "OpenSOAP/String.h"

using namespace OpenSOAP;
using namespace std;

std::string
createSOAPFaultMessage (std::string faultcodeStr, std::string faultstringStr,
			std::string faultactorStr, std::string detailStr)
{
    LIBXML_TEST_VERSION xmlKeepBlanksDefault(0);
    xmlDocPtr doc;
    xmlAttrPtr attr1, attr2;
    xmlNodePtr body, fault, faultcode, faultstring, faultactor, detail;
    xmlNsPtr soapEnvNs;
	
    /* make xml element */ 
    doc = xmlNewDoc((xmlChar*)("1.0"));
    attr1 = xmlNewDocProp(doc, (xmlChar*)("encoding"), 
			  (xmlChar*)("UTF-8")); /* ??? */
    attr2 = xmlNewDocProp(doc, (xmlChar*)("standalone"), 
			  (xmlChar*)("no")); /* ??? */
    
    /* make Envelope element */
    doc->children = xmlNewDocNode(doc, NULL, 
				  (xmlChar*)("Envelope"), NULL);
    xmlSetProp(doc->children, (xmlChar*)("encodingStyle"),
	       (xmlChar*)("http://schemas.xml.soap.org/soap/encoding/"));
    soapEnvNs = xmlNewNs(doc->children,
			 (xmlChar*)("http://schemas.xmlsoap.org/soap/envelope/"),
			 (xmlChar*)("SOAP-ENV"));
    xmlSetNs(doc->children, soapEnvNs);
    
    /* make Body element */
    body = xmlNewChild(doc->children, NULL, (xmlChar*)("Body"), NULL);
    xmlSetNs(body, soapEnvNs);
    
    /* make Fault element */
    fault = xmlNewChild(body, NULL, (xmlChar*)("Fault"), NULL);
    
    /* make faultcode element */
    faultcode = xmlNewChild(fault, NULL, (xmlChar*)("faultcode"),
			    (xmlChar*)faultcodeStr.c_str());
    
    /* make faultstring element */
    faultstring = xmlNewChild(fault, NULL, (xmlChar*)("faultstring"),
			      (xmlChar*)faultstringStr.c_str());

    /* make faultactor element */
    if (faultactorStr.length() >0) {
      faultactor = xmlNewChild(fault, NULL, (xmlChar*)("faultactor"),
			       (xmlChar*)faultactorStr.c_str());
    }
    
    /* make detail element */
    detail = xmlNewChild(fault, NULL, (xmlChar*)("detail"),
			 (xmlChar*)detailStr.c_str());
    
    /* output doc tree */
    xmlChar* mem_ptr;
    int size;
    xmlDocDumpMemory(doc, &mem_ptr, &size);
    
    std::string soapMsg ((char*)mem_ptr, size);
    
    /* free memory */
    xmlFreeDoc(doc);
    
    return soapMsg;
}

/*
  in: 
  ret: bool, exist operation
  throw: SOAPMsgIsInvalidException
 */
bool
#if 1
extractOperationAndParam (std::string soapMsg,
			  std::vector<std::string>& operationNames,
			  std::vector<std::string>& paramNames,
			  std::vector<std::string>& paramValues,
			  std::string* nsPtr) 
#else
extractOperationAndParam (std::string soapMsg,
			  std::vector<std::string>& operationNames,
			  std::vector<std::string>& paramNames,
			  std::vector<std::string>& paramValues) 
#endif
{
#ifdef DEBUG_
  cerr << "=====  extractOperationAndParam() begin =====" << endl;
#endif

#ifdef DEBUG_ 
  cerr << "soapMsg = [" << endl;
  cerr << soapMsg << endl;
  cerr << "]" << endl;
#endif
  
  bool operationExist = false;
  std::string query;
#if 0 // for debug ************************
	MsgAttrHandler msgAttrHndl;
	{
	for(int i=0; i<5; i++) {
		cerr << "XXXXXXXXX (" << i << ") XXXXXXXXXXX" << endl;
		msgAttrHndl.setMsg(soapMsg);
	}
	cerr << "XXXXXXX END XXXXXXXXX" << endl;
	}
#else
  MsgAttrHandler msgAttrHndl;
  msgAttrHndl.setMsg(soapMsg);
  //MsgAttrHandler msgAttrHndl(soapMsg);
#endif
  std::vector<std::string> values;
  std::vector<std::string>::iterator iter;
  
  operationNames.clear();
  paramNames.clear();
  paramValues.clear();  
  
  // get operation name
  query = "/Envelope/Body/?";
  if (0 > msgAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR: operation not exists ***" << endl;
    return operationExist;
  }
  
#ifdef DEBUG_
  cerr << "SOAPMessage::isItMsgID:values[0] = " << values[0] << endl;
#endif
  operationNames.push_back(values[0]);
  values.clear();

  // added 2002.03.11
  // get namespace
  if (nsPtr) {
    query = "/Envelope/Body/" + operationNames[0] +  ",xmlns=?";
    if (0 > msgAttrHndl.queryXml(query, values)) {
      cerr << "*** WARNING: no namespace ***" << endl;
    }
    else {
      *nsPtr = values[0];
#ifdef DEBUG_
      cerr << "############# SOAPMessage: namespace=[" 
	   << *nsPtr << "]:[" << values[0] << "]" << endl;
#endif
    }
    values.clear();
  }

  // get param name
  bool paramsExist = true;
  query = "/Envelope/Body/" + operationNames[0] +  "/??";
  if (0 > msgAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR: params not exists ***" << endl;
    return paramsExist;
  }
  
  if (paramsExist) {
    for (iter = values.begin(); iter != values.end(); iter++) {
#ifdef DEBUG_
      cerr << "SOAPMessage::params:values = " << *iter << endl;
#endif
      paramNames.push_back(*iter);    
    }
    values.clear();
    
    // get param values
    for (iter = paramNames.begin(); iter != paramNames.end(); iter++) {
      query = "/Envelope/Body/" + operationNames[0] +  "/"
			+ *iter + "=?";
      if (0 > msgAttrHndl.queryXml(query, values)) {
		cerr << "*** ERROR: no param value ***" << endl;
		return operationExist;
      }
      paramValues.push_back(values[0]);
      values.clear();
    }
  }// end of if (paramsExist)
  
  operationExist = true;  

#ifdef DEBUG_
  cerr << "=====  extractOperationAndParam() end =====" << endl;
#endif

  return operationExist;
}// end of extractOperationAndParam()

/*
  check Messaage ID
 */
bool
isItMsgID(const std::string soapMsg, bool& itIsMsgID)
{
#ifdef DEBUG_
  cerr << "=====  isItMsgID() begin =====" << endl;
#endif

  itIsMsgID = false;
  
  std::string query;
  MsgAttrHandler msgAttrHndl;
  msgAttrHndl.setMsg(soapMsg);
  //MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;

  query = "/Envelope/Header/opensoap-header-block/message_id=?";

  if (0 > msgAttrHndl.queryXml(query, values)) {
    return true;
  }

#ifdef DEBUG_
  cerr << "SOAPMessage::isItMsgID:values[0] = " << values[0] << endl;
#endif

  values.clear();
  
  query = "/Envelope/Body/?";
  if (0 > msgAttrHndl.queryXml(query, values)) {
    // if only <Body/> that mean messageID, return true;
    itIsMsgID = true;
    return true;
  }

#ifdef DEBUG_
  cerr << "SOAPMessage::isItMsgID:values[0] = " << values[0] << endl;
#endif

  if (values[0] == "") {
	// if has <message_id> and <Body> empty
	// it's messageID, return true
    itIsMsgID = true;
  }
#ifdef DEBUG_
  cerr << "=====  isItMsgID() end =====" << endl;  
#endif

  return true;
}

/*
 check sync or async
 */
bool
getAsyncInfoFromHeader(const std::string soapMsg,
		       bool& asyncHeaderExist, bool& isAsync)
{
#ifdef DEBUG_
  cerr << "=====  getAsyncInfoFromHeader() begin =====" << endl;
#endif

  asyncHeaderExist = false;
  isAsync = false;
  
  std::string query("/Envelope/Header/opensoap-header-block/async=?");
  MsgAttrHandler msgAttrHndl;
  msgAttrHndl.setMsg(soapMsg);
  //MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;
  if (0 > msgAttrHndl.queryXml(query, values)) {
    isAsync = false;
    return true;
  }

#ifdef DEBUG_
  cerr << "SOAPMessage::getAsyncInfoFromHeader:values[0] = " 
	<< values[0] << endl;
#endif

  if (values[0] == "true") {
    asyncHeaderExist = true;
    isAsync = true;
  }
  else if (values[0] == "false") {
    asyncHeaderExist = true;
    isAsync = false;
  }
  else {
    return true;
  }

#ifdef DEBUG_
  cerr << "=====  getAsyncInfoFromHeader() end =====" << endl;
#endif

  return true;
}// end of getAsyncInfoFromHeader()

/*
  check forward
 */
bool
getFwdInfoFromHeader(const std::string soapMsg,
		     bool& fwdPathExist, std::vector<std::string>& fwdPath,
		     bool& hopCountExist, unsigned int& hopCount)
{
#ifdef DEBUG_
  cerr << "=====  getFwdInfoFromHeader() begin =====" << endl;
#endif
  fwdPathExist = false;
  hopCountExist = false;
  fwdPath.clear();

  std::string query;
  MsgAttrHandler msgAttrHndl;
  msgAttrHndl.setMsg(soapMsg);
  //MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;
  std::vector<std::string>::iterator iter;

#if 1
  query = "/Envelope/Header/opensoap-header-block/forwarder/forward_path=??";  
  if (0 <= msgAttrHndl.queryXml(query, values)) {
    fwdPathExist = true;
    for (iter = values.begin(); iter != values.end(); iter++) {
      fwdPath.push_back(*iter);
    }
  }
  
  values.clear();
  query = "/Envelope/Header/opensoap-header-block/forwarder/hopcount=?";  
  if (0 <= msgAttrHndl.queryXml(query, values)) {
    hopCountExist = true;    
    hopCount = atoi(values[0].c_str());
  }
#endif

#ifdef DEBUG_
  cerr << "=====  getFwdInfoFromHeader() end =====" << endl;
#endif

  return true;
}// end of getFwdInfoFromHeader()

/*
   check Received Path
 */
bool
getReceivedPathInfo(const string soapMsg, 
		    bool& receivedPathExist, 
		    std::vector<std::string>& receivedPath) 
{
#ifdef DEBUG_
  cerr << "=====  getReceivedInfo() begin =====" << endl;
#endif

  receivedPathExist = false;
  receivedPath.clear();
  
  std::string query;
  MsgAttrHandler msgAttrHndl;
  msgAttrHndl.setMsg(soapMsg);
  //MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;
  std::vector<std::string>::iterator iter;
  
  query = "/Envelope/Header/opensoap-header-block/received_path/url=??";  
  if (0 <= msgAttrHndl.queryXml(query, values)) {
    receivedPathExist = true;
    for (iter = values.begin(); iter != values.end(); iter++) {
      receivedPath.push_back(*iter);
      cerr << "SOAPMessage::getReceivedInfo:values = " 
	   << *iter << endl;    
    }
  }

#ifdef DEBUG_
  cerr << "=====  getReceivedInfo() end =====" << endl;
#endif

  return true;
}

/*
  check TTL
 */
bool
getTTLInfo(const std::string soapMsg,
	   bool& ttlHeaderExist, unsigned int& ttl, std::string& ttlType)
{
#ifdef DEBUG_
  cerr << "=====  getTTLInfo() begin =====" << endl;
#endif

  ttlHeaderExist = false;
  
  std::string query;
  MsgAttrHandler msgAttrHndl;
  msgAttrHndl.setMsg(soapMsg);
  //MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;

  query = "/Envelope/Header/opensoap-header-block/ttl=?";
  if (0 > msgAttrHndl.queryXml(query, values)) {
	//edit 2002.03.18
    //ttl = 0;
    ttlType = "";
    ttlHeaderExist = false;
    return true;
  }
#ifdef DEBUG_
  cerr << "SOAPMessage::getTTLInfo: values[0] = " << values[0] << endl; 
#endif

  ttl = atoi(values[0].c_str());
  ttlHeaderExist = true;
  values.clear();
  
  query = "/Envelope/Header/opensoap-header-block/ttl,type=?";
  //query = "/Envelope/Header/opensoap-header-block/ttl=?";
  if (0 > msgAttrHndl.queryXml(query, values)) {
  //ttl = 0;
    ttlType = "";
    //ttlHeaderExist = false;
    return true;
  }

#ifdef DEBUG_
  cerr << "SOAPMessage::getTTLInfo: values[0] = " << values[0] << endl; 
#endif

  ttlType = values[0];

#ifdef DEBUG_
  cerr << "=====  getTTLInfo() end =====" << endl;
#endif

  return true;
}// end of getTTLInfo()


//create fault message for timeout
std::string
makeTimeoutMessage()
{
  
  std::string faultcodeStr("SOAP-ENV:Server.Timeout");
  std::string faultstringStr("Time Out occured in Server");
  std::string faultactorStr("");

  std::string serverName;
  size_t len = 128;  
  char* hostname = new char[len];
  int ret = gethostname(hostname, len);
  if (ret < 0) {
    serverName = "localhost";
  }
  else {
    hostent* hostEntPtr;
    hostEntPtr = gethostbyname(hostname);
#ifdef DEBUG_
    cerr << "hostEntPtr->h_name = " << hostEntPtr->h_name << endl;
#endif
    serverName = hostEntPtr->h_name;
  }
  delete[] hostname;
  
  std::string official_hostname = serverName;
  
  std::string detailStr = std::string("This FAULT is created by ") 
    + official_hostname + std::string(".");
  
  std::string soapMsgToIf
    = createSOAPFaultMessage(faultcodeStr, faultstringStr,
			     faultactorStr, detailStr);
  
  return soapMsgToIf;
}


//create fault message for operation not exist
std::string
makeOperationNotFoundMessage()
{
  std::string faultcodeStr("SOAP-ENV:Server.OperationNotFound");
  std::string faultstringStr("Request Operation is not Treated in Server");
  std::string faultactorStr("");
  
  std::string serverName;
  size_t len = 128;  
  char* hostname = new char[len];
  int ret = gethostname(hostname, len);
  if (ret < 0) {
    serverName = "localhost";
  }
  else {
    hostent* hostEntPtr;
    hostEntPtr = gethostbyname(hostname);
#ifdef DEBUG_
    cerr << "hostEntPtr->h_name = " << hostEntPtr->h_name << endl;
#endif
    serverName = hostEntPtr->h_name;
  }
  delete[] hostname;
  
  std::string official_hostname = serverName;
  
  std::string detailStr = std::string("This FAULT is created by ") 
    + official_hostname + std::string(".");
  
  std::string soapMsgToIf
    = createSOAPFaultMessage(faultcodeStr, faultstringStr,
			     faultactorStr, detailStr);
  return soapMsgToIf;
}

//create fault message for invalid request message
std::string
makeInvalidRequestMessage()
{
  std::string faultcodeStr("SOAP-ENV:Server.RequestInvalid");
  std::string faultstringStr("Request from Client is Invalid");
  std::string faultactorStr("");
  
  std::string serverName;
  size_t len = 128;  
  char* hostname = new char[len];
  int ret = gethostname(hostname, len);
  if (ret < 0) {
    serverName = "localhost";
  }
  else {
    hostent* hostEntPtr;
    hostEntPtr = gethostbyname(hostname);
#ifdef DEBUG_
    cerr << "hostEntPtr->h_name = " << hostEntPtr->h_name << endl;
#endif
    serverName = hostEntPtr->h_name;
  }
  delete[] hostname;
  
  std::string official_hostname = serverName;
  
  std::string detailStr = std::string("This FAULT is created by ") 
    + official_hostname + std::string(".");
  
  std::string soapMsgToIf
    = createSOAPFaultMessage(faultcodeStr, faultstringStr,
			     faultactorStr, detailStr);
  return soapMsgToIf;
}

//----- create Falut Messgage of No Entry -----//
std::string
makeNoEntryMessage()
{
  std::string faultcodeStr("SOAP-ENV:Server.NoEntry");
  std::string faultstringStr("No Corresponding Entry in Response Spool");
  std::string faultactorStr("");
  
  std::string serverName;
  size_t len = 128;  
  char* hostname = new char[len];
  int ret = gethostname(hostname, len);
  if (ret < 0) {
    serverName = "localhost";
  }
  else {
    hostent* hostEntPtr;
    hostEntPtr = gethostbyname(hostname);
#ifdef DEBUG_
    cerr << "hostEntPtr->h_name = " << hostEntPtr->h_name << endl;
#endif
    serverName = hostEntPtr->h_name;
  }
  delete[] hostname;
  
  std::string official_hostname = serverName;
  
  std::string detailStr = std::string("This FAULT is created by ") 
    + official_hostname + std::string(".");
  
  std::string soapMsgToIf
    = createSOAPFaultMessage(faultcodeStr, faultstringStr,
			     faultactorStr, detailStr);
  return soapMsgToIf;
}

//----- create Falut Messgage in the case of -----//
//----- Response from Serive is Empty        -----//
std::string
makeResponseIsEmptyMessage()
{
  std::string faultcodeStr("SOAP-ENV:Server.ResponseIsEmpty");
  std::string faultstringStr("Response from Service is Empty");
  std::string faultactorStr("");
  
  std::string serverName;
  size_t len = 128;  
  char* hostname = new char[len];
  int ret = gethostname(hostname, len);
  if (ret < 0) {
    serverName = "localhost";
  }
  else {
    hostent* hostEntPtr;
    hostEntPtr = gethostbyname(hostname);
#ifdef DEBUG_
    cerr << "hostEntPtr->h_name = " << hostEntPtr->h_name << endl;
#endif
    serverName = hostEntPtr->h_name;
  }
  delete[] hostname;
  
  std::string official_hostname = serverName;
  
  std::string detailStr = std::string("This FAULT is created by ") 
    + official_hostname + std::string(".");
  
  std::string soapMsgToIf
    = createSOAPFaultMessage(faultcodeStr, faultstringStr,
			     faultactorStr, detailStr);
  return soapMsgToIf;
}

//create fault message for limit size over
std::string
makeLimitSizeOverMessage()
{
  std::string faultcodeStr("SOAP-ENV:Server.LimitSizeOver");
  std::string faultstringStr("Message Size over Limit 1MB");
  std::string faultactorStr("");

  std::string serverName;
  size_t len = 128;  
  char* hostname = new char[len];
  int ret = gethostname(hostname, len);
  if (ret < 0) {
    serverName = "localhost";
  }
  else {
    hostent* hostEntPtr;
    hostEntPtr = gethostbyname(hostname);
#ifdef DEBUG_
    cerr << "hostEntPtr->h_name = " << hostEntPtr->h_name << endl;
#endif
    serverName = hostEntPtr->h_name;
  }
  delete[] hostname;
  
  std::string official_hostname = serverName;
  
  std::string detailStr = std::string("This FAULT is created by ") 
    + official_hostname + std::string(".");
  
  std::string soapMsgToIf
    = createSOAPFaultMessage(faultcodeStr, faultstringStr,
			     faultactorStr, detailStr);
  
  return soapMsgToIf;
}

//----- attach message ID into SOAP message -----//
//----- return message ID -----//
//----- use in MsgDrv -----//
std::string
attachMsgID(std::string& soapMsg, const std::string& fileIDOfMsgFromIf,
	    const std::string& operationName)
{
#ifdef DEBUG_
  cerr << "===== attachMsgID() begin =====" << endl;
#endif

  //messageID alread exist, not attach
  std::string msgIDContent;
  msgIDContent = getMsgIDContent(soapMsg);
  if (msgIDContent.size() > 0) {
    return msgIDContent;
  }
  
  std::string dot(".");
  
  std::string serverName = getLocalhostName();
  
  std::string receivedDateTime = fileIDOfMsgFromIf;
  
  msgIDContent = operationName + dot 
    + serverName + dot + receivedDateTime;
  
  // attach messageID
  xmlDocPtr soapMsgDoc = NULL;
  xmlNsPtr ns;
  xmlNodePtr node1, node2;
  
  /* build an XML tree from a std::string */
  
  char *soapMsgChar = new char[soapMsg.length()+1];
  memcpy (soapMsgChar, soapMsg.c_str(), soapMsg.length());
  soapMsgChar[soapMsg.length()] = 0x00;

  soapMsgDoc = xmlParseMemory(soapMsgChar, soapMsg.length());
  if (soapMsgDoc == NULL) {
//    fprintf(stderr, "parsing soapMsg =  '%s' failed\n", soapMsgChar);
	cerr << "attachMsgID: parsing soapMsg = [" << soapMsgChar
		<< "] failed." << endl;
    delete[] soapMsgChar;
    return NULL;
  }
  delete[] soapMsgChar;

  /* check document (root node) */
  node1 = xmlDocGetRootElement(soapMsgDoc);
  if (node1 == NULL) {
    //fprintf(stderr, "empty document\n");
	cerr << "attachMsgID: empty document" << endl;
    xmlFreeDoc(soapMsgDoc);
    return NULL;
  }
#ifdef DEBUG_
  cout << "node1->name = " << node1->name << endl;
#endif

  /* check namespace */
  ns = xmlSearchNsByHref(soapMsgDoc, node1,
			 (xmlChar*)("http://schemas.xmlsoap.org/soap/envelope/"));
  if (ns == NULL) {
    //fprintf(stderr, "wrong namespace of Envelope\n");
	cerr << "attachMsgID: wrong namespace of Envelope" << endl;
    xmlFreeDoc(soapMsgDoc);
    return NULL;
  }
  
  /* check document type (root node) */
  if (strcmp((const char*)(node1->name), "Envelope") != 0) {
    //fprintf(stderr, "wrong document type");
	cerr << "attachMsgID: wrong document type" << endl;
    xmlFreeDoc(soapMsgDoc);
    return NULL;
  }
  
  node2 = node1->xmlChildrenNode;
  if (node2 == NULL) {
    cerr << "node1->xmlChildrenNode == NULL" << endl;
    return NULL;
  }
#ifdef DEBUG_
  cerr << "node2->name = " << node2->name << endl;
#endif

  xmlNewChild(node2->xmlChildrenNode, NULL, (xmlChar*)("message_id"),
	      (xmlChar*)(msgIDContent.c_str()));

  //create message ID
  LIBXML_TEST_VERSION xmlKeepBlanksDefault(0);
  xmlDocPtr msgIDDoc;
  xmlAttrPtr attr1, attr2;
  xmlNodePtr body;
  xmlNsPtr soapEnvNs;
  
  /* make xml element */ 
  msgIDDoc = xmlNewDoc((xmlChar*)("1.0"));
  attr1 = xmlNewDocProp(msgIDDoc, (xmlChar*)("encoding"), 
			(xmlChar*)("UTF-8")); /* ??? */
  attr2 = xmlNewDocProp(msgIDDoc, (xmlChar*)("standalone"), 
			(xmlChar*)("no")); /* ??? */
  
  /* make Envelope element */
  msgIDDoc->children = xmlNewDocNode(msgIDDoc, NULL, 
				     (xmlChar*)("Envelope"), NULL);
  xmlSetProp(msgIDDoc->children, (xmlChar*)("encodingStyle"),
	     (xmlChar*)("http://schemas.xml.soap.org/soap/encoding/"));
  soapEnvNs = xmlNewNs(msgIDDoc->children,
		       (xmlChar*)("http://schemas.xmlsoap.org/soap/envelope/"),
		       (xmlChar*)("SOAP-ENV"));
  xmlSetNs(msgIDDoc->children, soapEnvNs);


  /* make Header element */
  xmlNodePtr header;
  header = xmlNewChild(msgIDDoc->children, NULL, (xmlChar*)("Header"), NULL);
  
  xmlNodePtr extBlock 
    = xmlNewChild(header, NULL, (xmlChar*)("opensoap-header-block"), NULL);
  xmlNsPtr opensoapExtNs 
    = xmlNewNs(extBlock, (xmlChar*)("http://header.opensoap.jp/1.0/"),
	       (xmlChar*)("opensoap-header"));
  xmlSetNs(extBlock, opensoapExtNs);
  
  xmlNewChild(extBlock, NULL, (xmlChar*)("message_id"),
	      (xmlChar*)(msgIDContent.c_str()));
  
/* make Body element */
  body = xmlNewChild(msgIDDoc->children, NULL, (xmlChar*)("Body"), NULL);
  xmlSetNs(body, soapEnvNs);
  
  /* output doc tree */
  xmlChar* mem_ptr;
  int size;
  
  xmlDocDumpMemory(soapMsgDoc, &mem_ptr, &size);
  soapMsg = std::string((char*)mem_ptr, size);

#ifdef DEBUG_
  cout << "SOAPMessage::attachMsgID:New soapMsg = " << soapMsg << endl;
#endif

  xmlDocDumpMemory(msgIDDoc, &mem_ptr, &size);
  std::string msgID = std::string((char*)mem_ptr, size);
  
  /* free memory */
  xmlFreeDoc(soapMsgDoc);
  xmlFreeDoc(msgIDDoc);

#ifdef DEBUG_
  cerr << "===== attachMsgID() end =====" << endl;
#endif

  return msgID;
}

//----- attach message ID into SOAP message -----//
//----- return message ID -----//
//----- use in SrvDrv -----//
std::string
attachOrUpdateMsgID(std::string& soapMsg, std::string msgIDContent)
{
#ifdef DEBUG_
  cerr << "===== attachOrUpdateMsgID() begin =====" << endl;
#endif

  XmlModifier xmlMod(soapMsg);
  XmlModifier::NameSpaceDef nsDef;
  std::string fmt;
  std::string val;
  int result;
  std::string newMsg;

  //if message ID alread exist, update messageID
  //else attach message ID
  std::string msgIDContentCurrent;
  msgIDContentCurrent = getMsgIDContent(soapMsg);
  if (msgIDContentCurrent.size() > 0) {
    
    
  }
  else {
    // add header 
    //fmt = "/Envelope/?";
    //val = "Header";
    fmt = "/Envelope/Header=?";
    val = "";
    //modified for Duplicate error : <Header>TAG
    //result = xmlMod.attach(fmt, val, NULL);
    result = xmlMod.attachNoDuplicate(fmt, val, NULL);
    if (result != 0) {
      cerr << "attach failed. code=(" << result << ")" << endl;
    }
    newMsg = xmlMod.toString();
#ifdef DEBUG_
    cout << "####################################" << endl;
    cout << "newMsg=[" << newMsg << "]" << endl;
    cout << "####################################" << endl;
#endif
    // add header block
    fmt = "/Envelope/Header/?";
    val = "opensoap-header-block";
    nsDef.href = "http://header.opensoap.jp/1.0/";
    nsDef.prefix = "opensoap-header";
    result = xmlMod.attach(fmt, val, &nsDef);
    if (result != 0) {
      cerr << "attach failed. code=(" << result << ")" << endl;
    }
    newMsg = xmlMod.toString();
#ifdef DEBUG_
    cout << "####################################" << endl;
    cout << "newMsg=[" << newMsg << "]" << endl;
    cout << "####################################" << endl;
#endif
    // add message ID
    fmt = "/Envelope/Header/opensoap-header-block/message_id=?";
    val = msgIDContent;
    result = xmlMod.attach(fmt, val, &nsDef);
    if (result != 0) {
      cerr << "attach failed. code=(" << result << ")" << endl;
    }
    newMsg = xmlMod.toString();
#ifdef DEBUG_
    cout << "####################################" << endl;
    cout << "newMsg=[" << newMsg << "]" << endl;
    cout << "####################################" << endl;
#endif
    soapMsg = newMsg;
  }

#ifdef DEBUG_
  cerr << "===== attachOrUpdateMsgID() end =====" << endl;
#endif

  return msgIDContent;
}


//----- attach backward path to response message,                 -----//
//----- if this server is the final receiver of async fwd message -----//
//----- This is used by SrvDrv -----//
bool
attachBackwardPathToResponse(std::string& msgFromService, 
			     std::string msgFromMsgDrv)
{
#ifdef DEBUG_
  cerr << "===== attachBackwardPathToResponse() begin =====" << endl;
#endif

  XmlModifier xmlMod(msgFromService);
  XmlModifier::NameSpaceDef nsDef;
  std::string fmt;
  std::string val;
  int result;
  std::string newMsg;
  
  std::string query;
  MsgAttrHandler msgAttrHndlRes;
  msgAttrHndlRes.setMsg(msgFromService);
  //MsgAttrHandler msgAttrHndlRes(msgFromService);
  MsgAttrHandler msgAttrHndlReq;
  msgAttrHndlReq.setMsg(msgFromMsgDrv);
  //MsgAttrHandler msgAttrHndlReq(msgFromMsgDrv);
  std::vector<std::string> values;
  
  // check if response has backward path
  values.clear();
  query = "/Envelope/Header/opensoap-header-block/backward_path=??";
  if (0 > msgAttrHndlRes.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    cerr << " query = [" << query << "] is not found."<< endl;
  }
  
  // if there is no backward path in Response message,
  // add backward path of request message,
  // otherwise do nothing
  if(values.size() > 0) {
    cerr << "Response Message already has <backward_path>." << endl;
    return true;
  }
  
  // retain backward_path from Request message
  values.clear();
  query = "/Envelope/Header/opensoap-header-block/backward_path=??";
  if (0 > msgAttrHndlReq.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    cerr << " query = [" << query << "] is not found."<< endl;
  }

  // return error, if no backward_path in Request message
  if (values.size() == 0) {
    cerr << "Request Message has no <backward_path>." << endl;
    return false;
  }
  
  // add backward_path
  fmt = "/Envelope/Header/opensoap-header-block/backward_path=?";
  for(std::vector<std::string>::iterator iter = values.begin();
      iter != values.end(); iter++) {
    val = *iter;
    result = xmlMod.attach(fmt, val, NULL);
    if (result != 0) {
      cerr << "attach failed. code=(" << result << ")" << endl;
      return false;
    }
  }// loop of for(iter)
  
  newMsg = xmlMod.toString();

#ifdef DEBUG_
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
#endif

  // add <response_msg> with value of "true"
  fmt = "/Envelope/Header/opensoap-header-block/response_msg=?";
  val = "true";
  result = xmlMod.attach(fmt, val, NULL);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
    return false;
  }
  
  newMsg = xmlMod.toString();

#ifdef DEBUG_
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
#endif

  msgFromService = newMsg;

#ifdef DEBUG_
  cerr << "===== attachBackwardPathToResponse() end =====" << endl;
#endif

  return true;
}// end of attachBackwardPathToResponse()

//----- attach TTL to response message,                 -----//
//----- if this server is the final receiver of async fwd message -----//
//----- This is used by SrvDrv -----//
bool
attachTTLToResponse(std::string& msgFromService, 
		    std::string msgFromMsgDrv)
{
#ifdef DEBUG_
  cerr << "===== attachTTLToResponse() begin =====" << endl;
#endif

  XmlModifier xmlMod(msgFromService);
  XmlModifier::NameSpaceDef nsDef;
  std::string fmt;
  std::string val;
  int result;
  std::string newMsg;
  
  std::string query;
  MsgAttrHandler msgAttrHndlRes;
  msgAttrHndlRes.setMsg(msgFromService);
  //MsgAttrHandler msgAttrHndlRes(msgFromService);
  MsgAttrHandler msgAttrHndlReq;
  msgAttrHndlReq.setMsg(msgFromMsgDrv);
  //MsgAttrHandler msgAttrHndlReq(msgFromMsgDrv);
  std::vector<std::string> values;
  
  // check if response has ttl
  values.clear();
  query = "/Envelope/Header/opensoap-header-block/ttl=?";
  if (0 > msgAttrHndlRes.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    cerr << " query = [" << query << "] is not found."<< endl;
  }
  
  // if there is no ttl path in Response message,
  // add backward path of request message,
  // otherwise do nothing
  if (values.size() > 0) {
    cerr << "Response Message already has <ttl>." << endl;
    return true;
  }
  
  // retain ttl from Request message
  values.clear();
  query = "/Envelope/Header/opensoap-header-block/ttl=?";
  if (0 > msgAttrHndlReq.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    cerr << " query = [" << query << "] is not found."<< endl;
  }
  
  // if no ttl in Request message, ttl becomes defalut value
  if (values.size() == 0) {
    cerr << "Request Message has no <ttl>." << endl;
    values[0] = 10;
  }
  
  // add ttl
  fmt = "/Envelope/Header/opensoap-header-block/ttl=?";
  val = values[0];
  result = xmlMod.attach(fmt, val, NULL);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
    return false;
  }
  
  newMsg = xmlMod.toString();

#ifdef DEBUG_
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
#endif

  // retain type of ttl from Request message
  values.clear();
  query = "/Envelope/Header/opensoap-header-block/ttl,type=?";
  if (0 > msgAttrHndlReq.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    cerr << " query = [" << query << "] is not found."<< endl;
  }
  
  // if no type of ttl in Request message, type of ttl becomes defalut value
  if (values.size() == 0) {
    cerr << "Request Message has no <ttl type=>." << endl;
    values[0] = "second";
  }
  
  // add type of ttl
  fmt = "/Envelope/Header/opensoap-header-block/ttl,type=?";
  val = values[0];
  result = xmlMod.attach(fmt, val);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
    return false;
  }
  
  newMsg = xmlMod.toString();

#ifdef DEBUG_
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
#endif

  msgFromService = newMsg;

#ifdef DEBUG_
  cerr << "===== attachTTLToResponse() end =====" << endl;
#endif

  return true;
}// end of attachTTLToResponse()

//----- attach received_path -----//
bool
attachReceivedPath(std::string& soapMsg, 
		   const std::string& fileIDOfMsgFromIf,
		   const std::string& url)
{
#ifdef DEBUG_
  cerr << "===== attachReceivedPath() begin =====" << endl;
#endif

  std::string receivedDateTime = fileIDOfMsgFromIf;
  
  //-----------------------------------------
  // XmlModifier
  //-----------------------------------------
  XmlModifier xmlMod(soapMsg);
  XmlModifier::NameSpaceDef nsDef;
  std::string fmt;
  std::string val;
  std::string newMsg;
  
  int result = 0;
  //add empty element
  //equal [fmt="/XXX/?" + val="AAA"] and [fmt="/XXX/AAA=?" + val=""]
  //add Namespace(ref. parent)
  //----------------------------------
  fmt = "/Envelope/Header/opensoap-header-block/?";
  val = "received_path";
  
  //namespace: if NULL , inherit parent NS
  result = xmlMod.attach(fmt, val, NULL);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
  }
  newMsg = xmlMod.toString();

#ifdef DEBUG_
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
#endif

  //----------------------------------
  //simple add
  //----------------------------------
  fmt = "/Envelope/Header/opensoap-header-block/received_path/url=?";
  val = url;
  
  result = xmlMod.attach(fmt, val);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
  }
  newMsg = xmlMod.toString();

#ifdef DEBUG_
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
#endif

  //----------------------------------
  //simple add
  //----------------------------------
  fmt = "/Envelope/Header/opensoap-header-block/received_path/time=?";
  val = fileIDOfMsgFromIf;
  
  result = xmlMod.attach(fmt, val);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
  }
  newMsg = xmlMod.toString();

#ifdef DEBUG_
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
#endif

  soapMsg = newMsg;

#ifdef DEBUG_
  cerr << "===== attachReceivedPath() end =====" << endl;
#endif

  return true;
}

//----- attach or update backward_path  -----//
bool
updateBackwardPath(std::string& soapMsg, 
		   const std::string& backwardURL)
{
#ifdef DEBUG_
  cerr << "===== updateBackwardPath() begin =====" << endl;
#endif

  //-----------------------------------------
  // XmlModifier
  //-----------------------------------------
  XmlModifier xmlMod(soapMsg);
  XmlModifier::NameSpaceDef nsDef;
  std::string fmt;
  std::string val;
  std::string newMsg;
  
  int result = 0;
  //----------------------------------
  //update
  //----------------------------------
  fmt = "/Envelope/Header/opensoap-header-block/backward_path=?";
  val = backwardURL;
  
  result = xmlMod.update(fmt, val);
  if (result != 0) {
    //if update target backward_path not exist, create and attach
#ifdef DEBUG_
    cerr << "attach failed. code=(" << result << ")" << endl;
#endif

    fmt = "/Envelope/Header/opensoap-header-block/backward_path=?";
    val = backwardURL;
    
    //append 3rd. argument for inherit parent Namespace 2003.01.08
    result = xmlMod.attach(fmt, val, NULL);
    if (result != 0) {
      cerr << "attach failed. code=(" << result << ")" << endl;
    }
  }
  
  newMsg = xmlMod.toString();

#ifdef DEBUG_
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
#endif

  soapMsg = newMsg;

#ifdef DEBUG_
  cerr << "===== updateBackwardPath() end =====" << endl;
#endif

  return true;
}// end of updateBackwardPath()

//----- Decrease value of <hopcount> in SOAP Message -----//
extern
bool
decreaseHopCount(std::string& soapMsg)
{
#ifdef DEBUG_
  cerr << "===== decreaseHopCount begin =====" << endl;
#endif

  std::string query;
  MsgAttrHandler msgAttrHndl;
  msgAttrHndl.setMsg(soapMsg);
  //MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;
  
  // check if response has ttl
  values.clear();
  query = "/Envelope/Header/opensoap-header-block/forwarder/hopcount=?";
  if (0 > msgAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    cerr << " query = [" << query << "] is not found."<< endl;
    return false;
  }
  
  int currentHopCount = atoi(values[0].c_str());
  
  //-----------------------------------------
  // XmlModifier
  //-----------------------------------------
  XmlModifier xmlMod(soapMsg);
  XmlModifier::NameSpaceDef nsDef;
  std::string fmt;
  std::string val;
  std::string newMsg;
  
  int result = 0;
  //----------------------------------
  //XV
  //----------------------------------
  fmt = "/Envelope/Header/opensoap-header-block/forwarder/hopcount=?";
  val = StringUtil::toString(currentHopCount - 1);
  
  result = xmlMod.update(fmt, val);
  if (result != 0) {
    return false;
  }
  
  newMsg = xmlMod.toString();

#ifdef DEBUG_
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
#endif

  soapMsg = newMsg;

#ifdef DEBUG_
  cerr << "===== decreaseHopCount end =====" << endl;
#endif

  return true;
}// end of decreaseHopCount


//attach message ID
//use in Service
bool
attachMsgID(std::string& soapMsg, const std::string& msgIDContent)
{
#ifdef DEBUG_
  cerr << "===== attachMsgID() begin =====" << endl;
#endif

  //attach message ID
  //-----------------------------------------
  // XmlModifier
  //-----------------------------------------
  XmlModifier xmlMod(soapMsg);
  XmlModifier::NameSpaceDef nsDef;
  std::string fmt;
  std::string val;
  std::string newMsg;
  
  int result = 0;
  //----------------------------------
  fmt = "/Envelope/?";
  val = "Header";
  

  result = xmlMod.attach(fmt, val, NULL);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
  }
  newMsg = xmlMod.toString();

#ifdef DEBUG_
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
#endif

  //----------------------------------
  // <opensoap-header:opensoap-header-block 
  //    xmlns:opensoap-header="http://header.opensoap.jp/1.0/" mustUnderstand="1">
  //----------------------------------
  //Namespace(new) add
  //----------------------------------
  
  fmt = "/Envelope/Header/?";
  val = "opensoap-header-block";
  nsDef.href = "http://header.opensoap.jp/1.0/";
  nsDef.prefix = "opensoap-header";
  
  result = xmlMod.attach(fmt, val, &nsDef);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
  }
  newMsg = xmlMod.toString();

#ifdef DEBUG_
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
#endif

  //----------------------------------
  // <opensoap-header:opensoap-header-block 
  //    xmlns:opensoap-header="http://header.opensoap.jp/1.0/" mustUnderstand="1">
  //----------------------------------
  //attributes add
  //----------------------------------
  fmt = "/Envelope/Header/opensoap-header-block,mustUnderstand=?";
  val = "1";
  
  result = xmlMod.attach(fmt, val);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
  }
  newMsg = xmlMod.toString();

#ifdef DEBUG_
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
#endif

  //----------------------------------
  // <opensoap-header:message_id>
  //   ServiceName.ServerName.RecieveDateTime
  // </opensoap-header:message_id>
  //----------------------------------
  //Namespace(ref) add
  //----------------------------------
  fmt = "/Envelope/Header/opensoap-header-block/message_id=?";
  val = msgIDContent;

  result = xmlMod.attach(fmt, val, NULL);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
  }
  newMsg = xmlMod.toString();

#ifdef DEBUG_
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
#endif

  //----------------------------------
  // <opensoap-header:response_msg>
  //   true
  // </opensoap-header:response_msg>
  //----------------------------------
  //Namespace(ref) add
  //----------------------------------
  fmt = "/Envelope/Header/opensoap-header-block/response_msg=?";
  val = "true";

  result = xmlMod.attach(fmt, val, NULL);
  if (result != 0) {
    cerr << "attach failed. code=(" << result << ")" << endl;
  }
  newMsg = xmlMod.toString();

#ifdef DEBUG_
  cout << "####################################" << endl;
  cout << "newMsg=[" << newMsg << "]" << endl;
  cout << "####################################" << endl;
#endif

  soapMsg = newMsg;

#ifdef DEBUG_
  cerr << "===== attachMsgID() end =====" << endl;
#endif

  return true;
}

//get messageID
std::string
getMsgIDContent(const std::string soapMsg)
{
  string msgIDContent("");

#ifdef DEBUG_
  cerr << "=====  getMsgIDContent() begin =====" << endl;
#endif

  std::string query;
  MsgAttrHandler msgAttrHndl;
  msgAttrHndl.setMsg(soapMsg);
  //MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;
  
  query = "/Envelope/Header/opensoap-header-block/message_id=?";
  if (0 > msgAttrHndl.queryXml(query, values)) {
    return msgIDContent;
  }

#ifdef DEBUG_
  cerr << "SOAPMessage::getMsgIDContent:values[0] = " << values[0] << endl;
#endif

  msgIDContent = values[0];

#ifdef DEBUG_
  cerr << "=====  getMsgIDContent() end =====" << endl;  
#endif

  return msgIDContent;
}

//check this server is final receiver from request message
// use in MsgDrv, AsyncSrvDrv
bool
isItAsyncFwdFinalReceiver(std::string soapMsg,
			  std::vector<std::string> soapServerEndPoints)
{
  bool itIsAsyncFwdFinalReceiver = false;
  //changed 2002/12/23
  //int  hopCountOfFinalReceiver = 1;
  int  hopCountOfFinalReceiver = 0;
#ifdef DEBUG_
  cerr << "=====  isItAsyncFwdFinalReceiver() begin =====" << endl;
#endif

#ifdef DEBUG_
  cerr << "soapMsg = [" << soapMsg << "]" << endl;
#endif

  std::string query;
  MsgAttrHandler msgAttrHndl;
  msgAttrHndl.setMsg(soapMsg);
  //MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;
  
  query = "/Envelope/Header/opensoap-header-block/forwarder/forward_path=??";
  if (0 > msgAttrHndl.queryXml(query, values)) {
#ifdef DEBUG_
    cerr << "*** ERROR ***" << endl;
    cerr << " query = [" << query << "] is not found."<< endl;
#endif
#if 1 //2002/12/24
    query = "/Envelope/Header/opensoap-header-block/received_path=?";
    if (0 > msgAttrHndl.queryXml(query, values)) {
      cerr << "*** ERROR ***" << endl;
      cerr << " query = [" << query << "] is not found."<< endl;
      return false;
      //return itIsAsyncFwdFinalReceiver;
    }
    else {
#ifdef DEBUG_
      cerr << "CHECK : received_path" << values[0] << endl;
#endif
      //itIsAsyncFwdFinalReceiver = true;
      return true;
    }
    
#else
    query = "/Envelope/Header/opensoap-header-block/forwarder/hopcount=?";
    if (0 > msgAttrHndl.queryXml(query, values)) {
      cerr << "*** ERROR ***" << endl;
      cerr << " query = [" << query << "] is not found."<< endl;
      return itIsAsyncFwdFinalReceiver;
    }
    else {
#ifdef DEBUG_
      cerr << "hopcount = " << values[0] << endl;
#endif
      if (atoi(values[0].c_str()) == hopCountOfFinalReceiver) {
	itIsAsyncFwdFinalReceiver = true;
      }
    }
    return itIsAsyncFwdFinalReceiver;
#endif

  }

#ifdef DEBUG_
  cerr << "SOAPMessage::isItAsyncFwdFinalReceiver:values.end() = " 
       << *(values.end() - 1) << endl;
#endif

#ifdef DEBUG_
  std::string soapServerEndPoint;
#endif
  for (std::vector<std::string>::const_iterator iter = soapServerEndPoints.begin();
       iter != soapServerEndPoints.end(); iter++) {
    if (*(values.end() - 1) == *iter) {
      itIsAsyncFwdFinalReceiver = true;
#ifdef DEBUG_
      soapServerEndPoint = *iter;
#endif
      break;
    }
  }
#ifdef DEBUG_
  cout << "soapServerEndPoint = " << soapServerEndPoint << endl;
  cout << "isItAsyncFwdFinalReceiver = " 
       << itIsAsyncFwdFinalReceiver << endl;;
#endif

#ifdef DEBUG_
  cerr << "=====  isItAsyncFwdFinalReceiver() end =====" << endl;  
#endif

  return itIsAsyncFwdFinalReceiver;
}// end of isItAsyncFwdFinalReceiver()

/*
  check backward_path
 */
bool
getBackwardPath(const std::string& soapMsg, std::string& backwardPath)
{
#ifdef DEBUG_
  cerr << "=====  getBackwardPath() begin =====" << endl;
#endif

  backwardPath = "";
  
  std::string query;
  MsgAttrHandler msgAttrHndl;
  msgAttrHndl.setMsg(soapMsg);
  //MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;

  query = "/Envelope/Header/opensoap-header-block/backward_path=?";
  if (0 > msgAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    cerr << "[" <<query << "] is not exist." << endl;
    return false;
  }

#ifdef DEBUG_
  cerr << "SOAPMessage::getTTLInfo: values[0] = " << values[0] << endl; 
#endif

  backwardPath = values[0];

#ifdef DEBUG_
  cerr << "=====  getBackwardPath() end   =====" << endl;
#endif

  return true;
}// end of getBackwardPath()

// check response message
// use in MsgDrv, Forwarder
bool
isItResponseMessage(std::string soapMsg)
{
  bool itIsResponseMessage = false;

#ifdef DEBUG_
  cerr << "=====  isItResponseMessage() begin =====" << endl;
  cerr << "soapMsg = [" << soapMsg << "]" << endl;
#endif

  std::string query;
  MsgAttrHandler msgAttrHndl;
  msgAttrHndl.setMsg(soapMsg);
  //MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;

  query = "/Envelope/Header/opensoap-header-block/response_msg=?";
  if (0 > msgAttrHndl.queryXml(query, values)) {
    return itIsResponseMessage;
  }

#ifdef DEBUG_
  cerr << "SOAPMessage::isItResponseMessage:values[0] = " 
       << values[0] << endl;
#endif

  if (values[0] == "true") {
    itIsResponseMessage = true;
  }

#ifdef DEBUG_
  cout << "isItResponseMessage = " 
       << itIsResponseMessage << endl;;
#endif

#ifdef DEBUG_
  cerr << "=====  isItResponseMessage() end =====" << endl;  
#endif

  return itIsResponseMessage;
}// end of isItResponseMessage()

//----- attach header for Service communication -----//
std::string
attachHeaderForService(std::string soapMsg)
{
  std::string newSOAPMsg;
  std::vector<std::string> operationNames;
  std::vector<std::string> paramNames;
  std::vector<std::string> paramValues;
  std::string endString("\r\n");
  extractOperationAndParam(soapMsg, operationNames, paramNames, paramValues);
  
  newSOAPMsg = std::string("POST ") + operationNames[0]
    + std::string(" HTTP/1.1") + endString;
  newSOAPMsg += std::string("Content-Type: text/xml; charset=UTF-8") 
    + endString;
  newSOAPMsg += std::string("Content-Length: ") 
    + StringUtil::toString(soapMsg.length()) + endString;
  newSOAPMsg += endString;
  newSOAPMsg += soapMsg;
  
  return newSOAPMsg;
  
}// end of attachHeaderForService(std::string soapMsg)

//----- remove header for Service communication -----//
std::string
removeHeaderForService(std::string recvMsg)
{
  char delim = '<';
  std::string::iterator openBraceLoc
    = find (recvMsg.begin(), recvMsg.end (), delim);
  if (openBraceLoc == recvMsg.end()) {
    cerr << "HTTP Message does not contain SOAP Message !" << endl;
    return std::string ("");
  }
  std::string soapMsg = std::string(openBraceLoc, recvMsg.end());
  
  return soapMsg;
}// end of removeHeaderForService (std::string& recvMsg)

//----- attach header SrvDrv communication -----//
std::string
attachHeaderForSrvDrv(std::string soapMsg)
{
  std::string newSOAPMsg;
  std::vector<std::string> operationNames;
  std::vector<std::string> paramNames;
  std::vector<std::string> paramValues;
  std::string endString("\r\n");
  extractOperationAndParam(soapMsg, operationNames, paramNames, paramValues);
  
  newSOAPMsg = std::string("HTTP/1.1 200 OK") + endString;
  newSOAPMsg += std::string("Content-Type: text/xml; charset=UTF-8") 
    + endString;
  newSOAPMsg += std::string("Content-Length: ") 
    + StringUtil::toString(soapMsg.length()) + endString;
  newSOAPMsg += endString;
  newSOAPMsg += soapMsg;
  
  return newSOAPMsg;
  
}// end of attachHeaderForService(std::string soapMsg)

//----- return hostname with FQDN -----//
std::string
getLocalhostName() 
{
  char localhostname[128];
  int ret0 = gethostname(localhostname, sizeof(localhostname));
  if (ret0 == 0) {

#ifdef DEBUG_
    cerr << "localhostname = " << localhostname << endl;
#endif

  }
  else {
    cerr << "gethostname() failed !!" << endl;
  }
  std::string officialHostname = localhostname;
  
  hostent* hostEntPtr;
  hostEntPtr = gethostbyname(localhostname);
  if (hostEntPtr != NULL) {

#ifdef DEBUG_
    cerr << "hostEntPtr = " << hostEntPtr << endl;
#endif

    if (hostEntPtr->h_name != NULL) {

#ifdef DEBUG_
      cerr << "hostEntPtr->h_name = " << hostEntPtr->h_name << endl;
#endif

      officialHostname = hostEntPtr->h_name;
    }
  }
  
  return officialHostname;
}

// check <undelete> tag
bool
hasUndeleteTag(const std::string soapMsg)
{
  bool hasUndel = false;

#ifdef DEBUG
  cerr << "=====  hasUndeleteTag() begin =====" << endl;
#endif

  std::string query;
  MsgAttrHandler msgAttrHndl(soapMsg);
  std::vector<std::string> values;
  std::vector<std::string>::iterator iter;

  query = "/Envelope/Header/opensoap-header-block/??";
  if (0 > msgAttrHndl.queryXml(query, values)) {
    //cerr << "*** ERROR ***: hasUndeleteTag" << endl;
    return hasUndel;
  }
  
  for (iter = values.begin(); iter != values.end(); iter++) {
	if (*iter == "undelete") {
		hasUndel = true;
		break;
	}
  }
#ifdef DEBUG
  cerr << "=====  hasUndeleteTag() end =====" << endl;  
#endif

  return hasUndel;
}

// convert enc
int
convertEncoding(std::string& soapMsg, 
		const std::string& fromEnc, 
		const std::string& toEnc)
{
  OpenSOAPByteArrayPtr fromStr;
  OpenSOAPByteArrayPtr toStr;
  const unsigned char* out = 0;
  int rc = OPENSOAP_NO_ERROR;

#ifdef DEBUG
  cerr << "convertEncoding: soapMsg=[" << soapMsg << "]" << endl;
#endif

  rc = OpenSOAPByteArrayCreateWithData((const unsigned char*)soapMsg.c_str(),
				       soapMsg.length()+1,
				       &fromStr);
  if (OPENSOAP_NO_ERROR != rc) {
    cerr << "convertEncoding: OpenSOAPByteArrayCreateWithData failed(" 
	 << rc << ")" << endl;
    return rc;
  }

  rc = OpenSOAPByteArrayCreate(&toStr);
  if (OPENSOAP_NO_ERROR != rc) {
    cerr << "convertEncoding: OpenSOAPByteArrayCreate failed(" 
	 << rc << ")" << endl;
    return rc;
  }
  
  rc = OpenSOAPStringConvertCharEncoding(fromEnc.c_str(),
					 fromStr,
					 toEnc.c_str(),
					 toStr);
  if (OPENSOAP_NO_ERROR != rc) {
    cerr << "convertEncoding: OpenSOAPStringConvertCharEncoding failed(" 
	 << rc << ")" << endl;
    return rc;
  }

  rc = OpenSOAPByteArrayBeginConst(toStr, &out);
  if (OPENSOAP_NO_ERROR != rc) {
    cerr << "convertEncoding: OpenSOAPByteArrayBeginConst failed(" 
	 << rc << ")" << endl;
    return rc;
  }

#ifdef DEBUG
  cerr << "ConvStr=[" << out << "]" << endl;
#endif

  soapMsg = (const char*)out;
  
  rc = OpenSOAPByteArrayRelease(fromStr);
  if (OPENSOAP_NO_ERROR != rc) {
    cerr << "convertEncoding: OpenSOAPByteArrayRelease failed(" 
	 << rc << ")" << endl;
    return rc;
  }

  // bugged anything...
  rc = OpenSOAPByteArrayRelease(toStr);
  if (OPENSOAP_NO_ERROR != rc) {
    cerr << "convertEncoding: OpenSOAPByteArrayRelease failed(" 
	 << rc << ")" << endl;
    return rc;
  }

  return OPENSOAP_NO_ERROR;
}

//added 2002/09/18
int
recvStdinCGI(std::string& soapMsg)
{
  int rc = OPENSOAP_NO_ERROR;

  //get CONTENT_LENGTH
  char* contentLengthEnv = NULL;
  contentLengthEnv = getenv("CONTENT_LENGTH");
  if (!contentLengthEnv) {
    cerr << "recvStdinCGI: getenv(CONTENT_LENGTH) failed." << endl;
    return -1;
  }
  unsigned char* readBuf = NULL;
  int contentLength = atoi(contentLengthEnv);
  if (contentLength <= 0) {
    cerr << "recvStdinCGI: getenv(CONTENT_LENGTH) invalid value." << endl;
    return -1;
  }
  else if (contentLength > OPENSOAP_MESSAGE_LIMIT_SIZE) {
    cerr << "recvStdinCGI: recv size limit over 1MB." << endl;
    return contentLength;
  }
  readBuf = new unsigned char[contentLength+1];
  memset(readBuf, 0x00, sizeof(readBuf));
#if defined(WIN32)
  int
#else
  ssize_t 
#endif
    readLen = 0;
  int remainLen = contentLength;
  while (remainLen > 0) {
#if defined(WIN32)
    readLen = _read(0, readBuf+(contentLength - remainLen), remainLen);
#else
    readLen = ::read(0, readBuf+(contentLength - remainLen), remainLen);
#endif
    if (0 == readLen) {
#ifdef DEBUG
      cerr << "recvStdinCGI::read 0 byte done." << endl;
#endif
      break;
    } 
    else if (0 > readLen) {
      cerr << "recvStdinCGI: read failed" << endl;
      delete[] readBuf;
      return -1;
    }
    remainLen -= readLen;

#ifdef DEBUG
    cerr << "recvStdinCGI: read:[" << readBuf 
	 << "] len=(" << readLen << ")" << endl;
#endif
  }

  OpenSOAPByteArrayPtr fromStr;
  rc = OpenSOAPByteArrayCreateWithData(readBuf,
				       contentLength,
				       &fromStr);
  if (OPENSOAP_NO_ERROR != rc) {
    cerr << "convertEncoding: OpenSOAPByteArrayCreateWithData failed(" 
	 << rc << ")" << endl;
    return rc;
  }

  string fromEnc;
  string toEnc;
  if (getEncodingCharset(soapMsg, fromEnc, toEnc, true)) {
#ifdef DEBUG
    cerr << "getEncodingCharset: fromEnc=[" << fromEnc << "]: toEnc=[" 
	 << toEnc << "]" << endl;
#endif

    // convert enc : FROM Content-Type: charset To UTF-8
    convertEncoding(fromStr, fromEnc, toEnc, soapMsg);
#ifdef DEBUG
    cerr << "soapMsg:[" << soapMsg << "]" << endl;
#endif

    replaceXmlEncoding(soapMsg, toEnc);
#ifdef DEBUG
    cerr << "replaceXmlEncoding:[" << soapMsg << "]" << endl;
#endif

    delete[] readBuf;

    rc = OpenSOAPByteArrayRelease(fromStr);
    if (OPENSOAP_NO_ERROR != rc) {
      cerr << "recvStdinCGI: OpenSOAPByteArrayRelease failed(" 
	   << rc << ")" << endl;
      return rc;
    }
    return contentLength;
  }
  else {
    delete[] readBuf;
    cerr << "getEncodingCharset failed" << endl;
    return -1;
  }
}

// convert enc
int
convertEncoding(OpenSOAPByteArrayPtr fromStr,
		const std::string& fromEnc, 
		const std::string& toEnc,
		std::string& soapMsg)
{
  OpenSOAPByteArrayPtr toStr;
  const unsigned char* out = 0;
  int rc = OPENSOAP_NO_ERROR;

  rc = OpenSOAPByteArrayCreate(&toStr);
  if (OPENSOAP_NO_ERROR != rc) {
    cerr << "convertEncoding: OpenSOAPByteArrayCreate failed(" 
	 << rc << ")" << endl;
    return rc;
  }

 rc = OpenSOAPStringConvertCharEncoding(fromEnc.c_str(),
					 fromStr,
					 toEnc.c_str(),
					 toStr);
  if (OPENSOAP_NO_ERROR != rc) {
    cerr << "convertEncoding: OpenSOAPStringConvertCharEncoding failed(" 
	 << rc << ")" << endl;
    return rc;
  }

  rc = OpenSOAPByteArrayBeginConst(toStr, &out);
  if (OPENSOAP_NO_ERROR != rc) {
    cerr << "convertEncoding: OpenSOAPByteArrayBeginConst failed(" 
	 << rc << ")" << endl;
    return rc;
  }

#ifdef DEBUG
  cerr << "ConvStr=[" << out << "]" << endl;
#endif

  soapMsg = (const char*)out;

  rc = OpenSOAPByteArrayRelease(toStr);
  if (OPENSOAP_NO_ERROR != rc) {
    cerr << "convertEncoding: OpenSOAPByteArrayRelease failed(" 
	 << rc << ")" << endl;
    return rc;
  }

  return OPENSOAP_NO_ERROR;
}

//get charset for convert enc
bool
getEncodingCharset(const string& soapMsg,
		   string& fromEnc, 
		   string& toEnc,
		   bool forCGI)
{
  const string DEFAULT_ENC("UTF-8");

  // check Content-Type: charset="?"
  const string CONTENT_TYPE_KEY("Content-Type:");
  const string CHARSET_KEY("charset=");
  const char dQuot = '"';
  const char xmlBeginKey = '<';
  //const char xmlFirstLineEndKey = '>';
  
  toEnc = DEFAULT_ENC;

  if (forCGI) {
    char* contentTypeEnv = 0;
    contentTypeEnv = getenv("CONTENT_TYPE");
    if (contentTypeEnv) {
      //parse charset
      string contentType(contentTypeEnv);
      int fromIdx = contentType.find(CHARSET_KEY);
      if (string::npos != fromIdx) {
	fromIdx += strlen(CHARSET_KEY.c_str());
	fromIdx += 1; // for "
	int toIdx = contentType.find(dQuot, fromIdx);
#ifdef DEBUG
	cerr << "From=(" << fromIdx << "): To=(" << toIdx << ")" << endl;
	cerr << "charset=[" << contentType.substr(fromIdx, (toIdx - fromIdx))
	     << "]" << endl;
#endif
	fromEnc = contentType.substr(fromIdx, (toIdx - fromIdx));
      }
    }
    else {
      //default from_enc=UTF-8
      fromEnc = DEFAULT_ENC;
    }
  }
  else {
    int xmlBeginIdx = soapMsg.find(xmlBeginKey);
    if (string::npos == xmlBeginIdx) {
      //empty xml
      return false;
    }

    //parse charset into HttpHeader
    int fromIdx = soapMsg.find(CONTENT_TYPE_KEY);
    int toIdx = string::npos;
    if (string::npos == fromIdx || xmlBeginIdx < fromIdx) {
      // default fromEnc = "UTF-8"
      fromEnc = DEFAULT_ENC;
    }
    else {
      fromIdx = soapMsg.find(CHARSET_KEY, fromIdx);
      if (string::npos == fromIdx || xmlBeginIdx < fromIdx) {
	// default fromEnc = "UTF-8"
	fromEnc = DEFAULT_ENC;
      }
      else {
	fromIdx += strlen(CHARSET_KEY.c_str());
	fromIdx += 1; // for "
	toIdx = soapMsg.find(dQuot, fromIdx);
#ifdef DEBUG
	cerr << "From=(" << fromIdx << "): To=(" << toIdx << ")" << endl;
	cerr << "charset=[" << soapMsg.substr(fromIdx, (toIdx - fromIdx))
	     << "]" << endl;
#endif
	fromEnc = soapMsg.substr(fromIdx, (toIdx - fromIdx));
      }
    }
  }
  return true;
}

//replace xml encoding charset
bool
replaceXmlEncoding(string& soapMsg, const string& toEnc)
{
  const string ENCODING_KEY("encoding=");
  const char dQuot = '"';
  const char xmlBeginKey = '<';
  const char xmlFirstLineEndKey = '>';

  int xmlBeginIdx = soapMsg.find(xmlBeginKey);
  if (string::npos == xmlBeginIdx) {
    //empty xml
    return false;
  }

  //parse xml encoding
  int xmlFirstLineEndIdx = soapMsg.find(xmlFirstLineEndKey, xmlBeginIdx);
  if (string::npos == xmlFirstLineEndIdx) {
    //error
    cerr 
      << "replaceXmlEncoding: Invalid xml message from Serivce!" 
      << endl;
    return false;
  }
  
  //rewrite encoding to "UTF-8"
  int fromIdx = soapMsg.find(ENCODING_KEY, xmlBeginIdx);
  if (string::npos != fromIdx && fromIdx < xmlFirstLineEndIdx) {
    fromIdx = soapMsg.find(dQuot, fromIdx);
    if (string::npos == fromIdx || xmlFirstLineEndIdx < fromIdx) {
      //no qQuot
      cerr 
	<< "replaceXmlEncoding: no double quotation marks in xml encoding"
	<< endl;
      return false;
    }
    fromIdx += 1;
    int toIdx = soapMsg.find(dQuot, fromIdx);
    soapMsg.replace(fromIdx, (toIdx - fromIdx), toEnc);
  }

  return true;
}
