/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: ServerCommon.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#if defined(WIN32)
#include <io.h>
#include <winsock.h>
#else
#include <unistd.h>
#endif

#include <iostream>

#include "ServerCommon.h"
#include "SrvConfAttrHandler.h"

using namespace OpenSOAP;
using namespace std;

//======== common function ==============

int OpenSOAP::read(int fd, std::string& data)
{
  char readBuf[4096];
  memset(readBuf, 0x00, sizeof(readBuf));
#if defined(WIN32)
  int
#else
  ssize_t 
#endif
	readLen = 0;
  data = "";

#if defined(WIN32)
  //readLen = _read(fd, readBuf, sizeof(readBuf)-1);
  readLen = recv(fd, readBuf, sizeof(readBuf), 0);
  if (readLen == SOCKET_ERROR) {
    cerr << TAG_ERR << "OpenSOAP::read failed" << endl;
    return -1;
  }
#else
  readLen = ::read(fd, readBuf, sizeof(readBuf)-1);
#endif

  if (0 == readLen) {
#ifdef DEBUG
    cerr << "OpenSOAP::read 0 byte done." << endl;
#endif
      
  } 
  else if (0 > readLen) {
    cerr << TAG_ERR << "OpenSOAP::read failed" << endl;
    return -1;
  }

  readBuf[readLen] = 0x00;
  data += readBuf;

#ifdef DEBUG
  cerr << "OpenSOAP::read: [" << readBuf 
       << "] len=(" << readLen << ")" << endl;
#endif

  //return 0;
  return data.length();
}

#if !defined(WIN32)
int OpenSOAP::read2(int fd, std::string& data)
{
  const int READBUFSIZE = 1049600;
  char* readBuf = new char[READBUFSIZE];

  memset(readBuf, 0x00, READBUFSIZE);
  ssize_t readLen = 0;
  data = "";
  
  while (1) {
    //cerr << "********** NOW READ START-->" << endl;
    readLen = ::read(fd, readBuf, sizeof(readBuf)-1);

    if (0 == readLen) {
#ifdef DEBUG
      cerr << "OpenSOAP::read 0 byte done." << endl;
#endif
      break;
    } 
    else if (0 > readLen) {
      cerr << TAG_ERR << "OpenSOAP::read failed" << endl;
      delete[] readBuf;
      return -1;
    }

    readBuf[readLen] = 0x00;
    data += readBuf;

#ifdef DEBUG
    cerr << "OpenSOAP::read2: [" << readBuf 
	 << "] len=(" << readLen << ")" << endl;
#endif
  }

  delete[] readBuf;
  
  return data.length();
}
#endif


int OpenSOAP::write(int fd, const std::string& data)
{
  char writeBuf[4096];
  memset(writeBuf, 0x00, sizeof(writeBuf));
#if defined(WIN32)
	int
#else
  ssize_t 
#endif
		writeLen = 0;

  //std::string::size_type remainLen = data.length();
  int remainLen = data.length();

  while (remainLen > 0) {
#if defined(WIN32)
    writeLen = send(fd, 
		    data.c_str()+(data.length() - remainLen),
		    remainLen,
		    0);
    if (writeLen == SOCKET_ERROR) {
      cerr << TAG_ERR << "OpenSOAP::write failed" << endl;
      return -1;
    } 
#else
    writeLen = ::write(fd, 
		       data.c_str()+(data.length() - remainLen), 
		       remainLen);
#endif
    if (0 > writeLen) {
      cerr << TAG_ERR << "OpenSOAP::write failed" << endl;
      return -1;
    } 
    else if (writeLen > remainLen) {
      cerr << "*** W ***:" << "OpenSOAP::write: invalid write length." << endl;
    }
    
    remainLen -= writeLen;
  }

  return data.length() - remainLen;
}



// End of ServerCommon.cpp
