/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: SrvConf.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif /* HAVE_CONFIG_H */

#ifndef SYSCONFDIR
#if defined(WIN32)
# define SYSCONFDIR "/opensoap/etc/"
#else
# define SYSCONFDIR "/usr/local/opensoap/etc/"
#endif
#endif /* SYSCONFDIR */

#if defined(WIN32)
//#include <process.h>
#else
/*
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>

#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>

#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
*/
#endif

#include <stdexcept>

#include <string>
#include <iostream>

/* libxml2 */
#if defined(HAVE_LIBXML2_XMLVERSION_H)
#include <libxml2/tree.h>
#include <libxml2/parser.h>
#else
#include <libxml/tree.h>
#include <libxml/parser.h>
#endif /* !HAVE_LIBXML2_XMLVERSION_H */

#include "ServerCommon.h"
#include "SSMLAttrMgrProtocol.h"
#include "XmlQuery.h"
#include "SrvConf.h"
#include "StringUtil.h"

using namespace OpenSOAP;
using namespace std;

const std::string SrvConf::SERVER_CONF_FILE = "server.conf";

// for LOG
static const std::string CLASS_SIG = "SrvConf";

//XMLt@Cw肷
SrvConf::SrvConf()
  : xmlFilePath_(SYSCONFDIR)
  , confDocPtr_(NULL)
{
  //fBNgpX/ŏI悤ɂĂ
  if (FILE_PATH_DELMIT != xmlFilePath_.at(xmlFilePath_.length() - 1)) {
    xmlFilePath_ += FILE_PATH_DELMIT;
  }

#ifdef DEBUG
  cerr << CLASS_SIG << "::" << CLASS_SIG << ": xmlFilePath_=["
       << xmlFilePath_ << "]" << endl;
#endif /* DEBUG */

  //server.conft@C̃[h
  if (!loadXml()) {
    cerr << TAG_ERR << CLASS_SIG << "::" << CLASS_SIG 
	 << ": load server.conf file failed" << endl;
    throw std::exception();
  }
}

SrvConf::~SrvConf()
{
  //DOM\J
  if (confDocPtr_) {
    xmlFreeDoc(confDocPtr_);
  }
}

bool
SrvConf::loadXml()
{
  if (confDocPtr_) {
    xmlFreeDoc(confDocPtr_);
  }
  /* parse ssml begin */
  /* COMPAT: Do not generate nodes for formatting spaces */
  LIBXML_TEST_VERSION
    xmlKeepBlanksDefault(0);
  xmlNodePtr node1 = NULL;
  
  /* build an XML tree from a file */
  std::string srvConfFilePath;
  srvConfFilePath = xmlFilePath_ + SERVER_CONF_FILE;
  
  confDocPtr_ = xmlParseFile((const char*)(srvConfFilePath.c_str()));
  if (confDocPtr_ == NULL) {
    cerr << TAG_WRN << CLASS_SIG << "::addXmlList: "
	 << "parsing file " << srvConfFilePath << " failed." << endl;
    return false;
  }
  
  /* check document (root node) */
  node1 = xmlDocGetRootElement(confDocPtr_);
  if (node1 == NULL) {
    cerr << TAG_WRN << CLASS_SIG << "::addXmlList: "
	 << "empty document " << srvConfFilePath << endl;
    xmlFreeDoc(confDocPtr_);
    return false;
  }

#ifdef DEBUG
  cerr << CLASS_SIG << "::addXmlList: add doc." << endl;
#endif /* DEBUG */
  
  return true;
}

int
SrvConf::query(const std::string& queryStr, vector<string>& values) 
{
#ifdef DEBUG
  cerr << CLASS_SIG << "::query() arg=[" << queryStr << "]" << endl;
#endif /* DEBUG */

  XmlQuery* xmlQuery = new XmlQuery(queryStr);
  
  //std::vector<std::string> values;
  xmlNodePtr node = xmlDocGetRootElement(confDocPtr_);

  int ret = xmlQuery->getValue(confDocPtr_, node, values);

#ifdef DEBUG
  cerr << "xmlQuery ret=(" << ret << ")" << endl;
#endif /* DEBUG */
  
  delete xmlQuery;

  if (ret == 0) {
    ret = values.size();
  }
  return ret;
}

string 
SrvConf::getLogPath()
{
  string qry = "/server_conf/log/path=?";
  vector<string> values;
  int ret = query(qry, values);
  string retStr;
  if (ret > 0) {
    retStr = values[0];
    if (!retStr.empty() && retStr[retStr.length()-1] != FILE_PATH_DELMIT) {
      retStr += FILE_PATH_DELMIT;
    }
  }
  return retStr;
}

string 
SrvConf::getSoapMessageSpoolPath()
{
  string qry = "/server_conf/spool/soap_message/path=?";
  vector<string> values;
  int ret = query(qry, values);
  string retStr;
  if (ret > 0) {
    retStr = values[0];
    if (!retStr.empty() && retStr[retStr.length()-1] != FILE_PATH_DELMIT) {
      retStr += FILE_PATH_DELMIT;
    }
  }
  return retStr;
}

string 
SrvConf::getAsyncTablePath()
{
  string qry = "/server_conf/spool/async_table/path=?";
  vector<string> values;
  int ret = query(qry, values);
  string retStr;
  if (ret > 0) {
    retStr = values[0];
    if (!retStr.empty() && retStr[retStr.length()-1] != FILE_PATH_DELMIT) {
      retStr += FILE_PATH_DELMIT;
    }
  }
  return retStr;
}

string 
SrvConf::getPIDPath()
{
  string qry = "/server_conf/run/pid/path=?";
  vector<string> values;
  int ret = query(qry, values);
  string retStr;
  if (ret > 0) {
    retStr = values[0];
    if (!retStr.empty() && retStr[retStr.length()-1] != FILE_PATH_DELMIT) {
      retStr += FILE_PATH_DELMIT;
    }
  }
  return retStr;
}

string 
SrvConf::getSocketPath()
{
  string qry = "/server_conf/run/socket/path=?";
  vector<string> values;
  int ret = query(qry, values);
  string retStr;
  if (ret > 0) {
    retStr = values[0];
    if (!retStr.empty() && retStr[retStr.length()-1] != FILE_PATH_DELMIT) {
      retStr += FILE_PATH_DELMIT;
    }
  }
  return retStr;
}

string 
SrvConf::getSecKeyPath()
{
  string qry = "/server_conf/security/keys/path=?";
  vector<string> values;
  int ret = query(qry, values);
  string retStr;
  if (ret > 0) {
    retStr = values[0];
    if (!retStr.empty() && retStr[retStr.length()-1] != FILE_PATH_DELMIT) {
      retStr += FILE_PATH_DELMIT;
    }
  }
  return retStr;
}

string 
SrvConf::getSSMLPath()
{
  string qry = "/server_conf/ssml/path=?";
  vector<string> values;
  int ret = query(qry, values);
  string retStr;
  if (ret > 0) {
    retStr = values[0];
    if (!retStr.empty() && retStr[retStr.length()-1] != FILE_PATH_DELMIT) {
      retStr += FILE_PATH_DELMIT;
    }
  }
  return retStr;
}

string 
SrvConf::getBackwardProtocol()
{
  string qry = "/server_conf/backward/protocol=?";
  vector<string> values;
  int ret = query(qry, values);
  string retStr;
  if (ret > 0) {
    retStr = values[0];
  }
  return retStr;
}

vector<string> 
SrvConf::getBackwardHostnames()
{
  string qry = "/server_conf/backward/hostname=??";
  vector<string> values;
  int ret = query(qry, values);
  return values;
}

string 
SrvConf::getBackwardPort()
{
  string qry = "/server_conf/backward/port=?";
  vector<string> values;
  int ret = query(qry, values);
  string retStr;
  if (ret > 0) {
    retStr = values[0];
  }
  return retStr;
}

string 
SrvConf::getBackwardCgiLocation()
{
  string qry = "/server_conf/backward/cgi_location=?";
  vector<string> values;
  int ret = query(qry, values);
  string retStr;
  if (ret > 0) {
    retStr = values[0];
  }
  return retStr;
}

string 
SrvConf::getForwarderProtocol()
{
  string qry = "/server_conf/forwarder/protocol=?";
  vector<string> values;
  int ret = query(qry, values);
  string retStr;
  if (ret > 0) {
    retStr = values[0];
  }
  return retStr;
}

string 
SrvConf::getForwarderHostname()
{
  string qry = "/server_conf/forwarder/hostname=?";
  vector<string> values;
  int ret = query(qry, values);
  string retStr;
  if (ret > 0) {
    retStr = values[0];
  }
  return retStr;
}

string 
SrvConf::getForwarderPort()
{
  string qry = "/server_conf/forwarder/port=?";
  vector<string> values;
  int ret = query(qry, values);
  string retStr;
  if (ret > 0) {
    retStr = values[0];
  }
  return retStr;
}

string 
SrvConf::getForwarderCgiLocation()
{
  string qry = "/server_conf/forwarder/cgi_location=?";
  vector<string> values;
  int ret = query(qry, values);
  string retStr;
  if (ret > 0) {
    retStr = values[0];
  }
  return retStr;
}

int
SrvConf::getForwarderTimeout()
{
  string qry = "/server_conf/forwarder/timeout=?";
  vector<string> values;
  int ret = query(qry, values);
  string retStr;
  int retVal = 0;
  if (ret > 0) {
    StringUtil::fromString(values[0], retVal);
  }
  return retVal;
}

string
SrvConf::getAddSignature()
{
  string qry = "/server_conf/add_signature=?";
  vector<string> values;
  int ret = query(qry, values);
  string retStr;
  if (ret > 0) {
    retStr = values[0];
  }
  return retStr;
}

bool 
SrvConf::isAddSignatureTrue()
{
  if (getAddSignature() == "true") {
    return true;
  }
  else {
    return false;
  }
}

