/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: SrvConfAttrMgr_main.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#if defined(WIN32)
#include <windows.h>
#include <io.h>
#else
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#endif

#include <string>
#include <fstream>
#if !defined(NO_LOGGER)
#include "Logger.h"
#endif
#include "ServerCommon.h"
#include "XmlQuery.h"
#include "SrvConfAttrMgr.h"
#include "PIDFactory.h"

using namespace OpenSOAP;
using namespace std;

int main(int argc, char** argv)
{
  SrvConfAttrMgr srvConfAttrMgr;

#if defined(NO_LOGGER)

	string logFile = srvConfAttrMgr.getLogPath() + "srvConfAttrMgr.log";

	FILE* outfp = fopen(logFile.c_str(), "w");
	if (!outfp) {
		return 1;
	}
	if (argc <=1) {
		fprintf(outfp, "invalid initial argument. argc=(%d)\n", argc);
		fprintf(outfp, "exit.\n");
		fclose(outfp);
		return EXIT_FAILURE;
	}

	int old;
	int olderr;
	//backup stdout
	old = _dup(1);
	//stdout -> outfp
	if (-1 == _dup2(_fileno(outfp), 1)) {
		fprintf(outfp, "_dup2 to stdout failed.\n");
		fprintf(outfp, "exit.\n");
		fclose(outfp);
		return EXIT_FAILURE;
	}
	//backup stderr
	olderr = _dup(2);
	//stderr -> outfp
	if (-1 == _dup2(_fileno(outfp), 2)) {
		fprintf(outfp, "_dup2 to stderr failed.\n");
		fprintf(outfp, "exit.\n");
		fclose(outfp);
		return EXIT_FAILURE;
	}
#else
  // set log level  
  Logger::InitOpt opt;
  opt.outputLevel = LOG_DEBUG;
  Logger::init(opt);
  //write start message
  Logger::printLog(NULL, " SrvConfAttrMgr_main start", LOG_INFO,
		   0x00100003L, "exe");
#endif

#if !defined(WIN32)
  // make it to daemon & write pid file
#if HAVE_DAEMON
  daemon(0, 1);
#endif

  PIDFactory* pidFactory = new PIDFactory();
  pidFactory->makePID(SRVCONFATTRMGR_PID_FILE);
  delete pidFactory;

#endif

//  SrvConfAttrMgr srvConfAttrMgr;
  srvConfAttrMgr.run();

#if defined(NO_LOGGER)
	fclose(outfp);
	_dup2(old, 1);
	_dup2(olderr, 2);
#else
  //write stop message
  Logger::printLog(NULL, " SrvConfAttrMgr_main stop", LOG_INFO,
		   0x00100003L, "exe");
  Logger::term();
#endif

  return 0;
}

