/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: SrvDrv.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#if defined(WIN32)
#include <process.h>
#else
#include <sys/time.h>
#include <unistd.h>
#endif

#include <iostream>
#include <string>
#include <vector>

/* SSML Information Function library */
#include "SSMLInfoFunc.h"
#include "SSMLInfo.h"

/* server common library */
#include "ServerCommon.h"
#include "ChannelDescriptor.h"
//#include "FIFOSrvChnlSelector.h"
#include "ResSpoolPushChnlSelector.h"
#include "FwdQueuePushChnlSelector.h"
#include "FwdInfoFunc.h"
#include "SignatureFunc.h"

/* srvdrv library */
#include "SrvDrv.h"

/* common library */
#include "connection.h"
#include "FileIDFunc.h"
#include "SOAPMessage.h"
#include "StringUtil.h"
#include "Timeout.h"

using namespace OpenSOAP;
using namespace std;

#define DEBUG

//windowspmutexhcJEg
#if defined(WIN32)
int SrvDrv::mtCount = 0;
#endif



//----- Constructor -----//
SrvDrv::SrvDrv(const ChannelDescriptor& chnlDesc)
  : chnlDesc_(chnlDesc)
{
#ifdef DEBUG
  cerr << "SrvDrv::SrvDrv" << endl;
#endif

#if defined(WIN32)
	if (mtCount > 10000) {
		mtCount = 0;
	}
	std::string mtId = "Spool" + StringUtil::toString(mtCount++);
	_mt = CreateMutex(NULL, FALSE, mtId.c_str());
#else
  pthread_mutex_init(&_mt, NULL);
#endif
  downside_timeout = 5;
  _isAsync = false;

  bool localURLExist = getBackwardURLs(localURLs);
  if (!localURLExist) {
    cerr << "local URL not Exist" << endl;
  }
  
}// end of SrvDrv ()

//----- Destructor -----//
SrvDrv::~SrvDrv ()
{
#ifdef DEBUB
  cerr << "SrvDrv::~SrvDrv" << endl;
#endif
	//Mutext̔j
#if defined(WIN32)
	CloseHandle(_mt);
#else
  pthread_mutex_destroy(&_mt);
#endif
}// end of ~SrvDrv ()

//----- [J̓Iy[Vs -----//
bool
SrvDrv::execLocalSyncOperation()
{
#if !defined(WIN32)
  struct timeval tv_start, tv_end;
  struct timezone tz;
  
  gettimeofday(&tv_start, &tz);
  cerr << "tv_start.tv_sec  = " << tv_start.tv_sec << endl;
  cerr.flush();
#endif

#ifdef DEBUG
  // for debug
  cerr << "SrvDrv::execLocalSyncOperation() ttl=(" << ttl << ")" << endl;
  // for debug
  cerr << "SrvDrv::execLocalSyncOperation() create thread..." << endl;
#endif

  //ړȈsȂXbh𐶐

	// SSML蓖YIy[V̒ʐM^Cv
	// fAɂ쐬SrvDrṽCX^X
	// U蕪
	// added 2002.03.11
	string connectionType = ssmlInfo.getConnectionMethod();

#ifdef DEBUG
  cerr << "SrvDrv::execLocalSyncOperation(): connectionType=["
       << connectionType << "]" << endl;
#endif


#if defined(WIN32)
	HANDLE thread_t;
	void (*funcPtr)(void*);
#else
	pthread_t thread_t;
	void* (*funcPtr)(void*);
#endif

	if (SRV_CONNCT_SOCKET == connectionType) {
		funcPtr = localSyncThrSocket;
	}
	else if (SRV_CONNCT_STDIO == connectionType) {
		funcPtr = localSyncThrStdio;
	}
	else if (SRV_CONNCT_FIFO == connectionType ) {
		//
	}
	else if (SRV_CONNCT_IPC == connectionType ) {
		//
	}
	else if (SRV_CONNCT_MODULE == connectionType ) {
		//
	}
	else if (SRV_CONNCT_COM == connectionType ) {
		//
	}
	else {
		cerr << "SrvDrv::execLocalSyncOperation(): "
			<< "service connection type error." << endl;
      return false;
	}

#if defined(WIN32)
	thread_t = (HANDLE)_beginthread(funcPtr, 0, this);
#else
  int status = pthread_create(&thread_t, NULL, funcPtr, this);
  
#ifdef DEBUG
  // for debug
  cerr << "SrvDrv::execLocalSyncOperation() "
       << "thread create done... status=(" << status << ")" << endl;
#endif

#endif

  //^C}[IuWFNg
  OpenSOAP::Timeout timeout(ttl, thread_t);
  //^CAEgĎJn
  timeout.start();
  
  //ړȈɃXbh
#if defined(WIN32)
  if (WaitForSingleObject(thread_t, INFINITE) == STATUS_WAIT_0) {
#ifdef DEBUG
    cerr << "SrvDrv::execLocalSyncOperation() thread joined" 
	 << endl;
#endif
  }
  else {
    cerr 
      << "SrvDrv::execLocalSyncOperation() thread join error"
      << endl;
  }
#else
  status = pthread_join(thread_t, NULL);
  
  if (status != 0) {
    cerr << "SrvDrv::execLocalSyncOperation() pthread join error" << endl;
  }
#endif

  //ړIIɂ^C}[~
  timeout.finish();
  
#ifdef DEBUG
  // for debug
  cerr << "SrvDrv::execLocalSyncOperation() bye..." << endl;
  cerr << "timeout.targetThreadIsCanceled = " 
       << timeout.targetThreadIsCanceled() << endl;
#endif

#if !defined(WIN32)
  gettimeofday(&tv_end, &tz);
  cerr << "tv_end.tv_sec  = " << tv_end.tv_sec << endl;
  cerr.flush();
#endif

  return timeout.targetThreadIsCanceled();
}

//----- [JœIy[Vs߂̃Xbh  -----//
//----- Socket -----
#if defined(WIN32)
void
#else
void*
#endif
SrvDrv::localSyncThrSocket (void* arg) 
{
  SrvDrv* that = (SrvDrv*)arg;
  

#ifdef DEBUG
  // for debug  
  cerr << "SrvDrv::localSyncThrFunc: thr_func wait..." << endl;
#endif

  std::string msgFromService;
  
  // T[rX̕ԓ҂
  std::string hostname = that->ssmlInfo.getHostname();
  unsigned int port = that->ssmlInfo.getPort();
  
#ifdef DEBUG
  std::cerr << "SrvDrv::localSyncThrSocket: "
	    << "hostname = " << hostname << std::endl;
  std::cerr << "SrvDrv::localSyncThrSocket: "
	    << "port = " << port << std::endl;
#endif

  int fd = openSocket(hostname, port);
#ifdef DEBUG
  std::cerr << "SrvDrv::localSyncThrSocket: "
	    << "fd = " << fd << endl;
#endif

  if (0 > fd) {
    cerr << "SrvDrv::localSyncThrSocket: socket open failed" << endl;
    msgFromService 
      = createSOAPFaultMessage("Service.ConnectionNotEstablished",
			       "connection not established",
			       "server",
			       that->ssmlInfo.getOperationName() 
			       + " not respond");
    that->fileIDOfMsgFromService 
      = convertToFileID(msgFromService, 
			that->srvConf.getSoapMessageSpoolPath());
#if defined(WIN32)
	_endthread();
#else
	pthread_exit(NULL);
    return NULL;
#endif
  }
  
  #if defined(WIN32)
	WaitForSingleObject(that->_mt, INFINITE);
#else
int lock_status = pthread_mutex_lock (&(that->_mt));
  if (0 != lock_status) {
    cerr << "[[Error!!]] void* SrvDrv::localSyncThrSocket(void*)" << endl;
    cerr << "            pthread_mutex_lock() failed." << endl;
    cerr.flush();
    return NULL;
  }
#endif

  // revert file contents from file ID
  std::string msgFromMsgDrv 
    = revertFromFileID(that->fileIDOfMsgFromMsgdrv, 
		       that->srvConf.getSoapMessageSpoolPath());

  // add signature if needed
  if (that->srvConf.isAddSignatureTrue()) {
    if (!addSignatureToString(msgFromMsgDrv,
			      that->srvConf.getSecKeyPath() 
			      + SECURITY_DATA_PRIV_KEY)){
      cerr << "[[Error!!]] void* MsgDrv::localSyncThrSocket(void*)" << endl;
      cerr << "            addSignature failed." << endl;
    };
  }
  
  // attach HTTP header for service
  std::string msgToService = attachHeaderForService(msgFromMsgDrv);
  
  // send to service
  if (0 > sendSocket(fd, msgToService)) {
    cerr << "SrvDrv::localSyncThrSocket "
	 << "write failed to InetSoeketService" << endl;
	  // Mutex̉
#if defined(WIN32)
		ReleaseMutex(that->_mt);
		_endthread();
#else
	  int unlock_status = pthread_mutex_unlock (&(that->_mt));
	  if (0 != unlock_status) {
	    cerr << "[[Error!!]] void* SrvDrv::localSyncThrSocket(void*)" 
			<< endl;
		cerr << "            pthread_mutex_unlock() failed." << endl;
	    cerr.flush();
		pthread_exit(NULL);
	    return NULL;
	  }
	pthread_exit(NULL);
    return NULL;
#endif
  }

#ifdef DEBUG
  cerr << "** SrvDrv ** WRITE TO InetSocketSrv=[" 
       << msgToService << "]" << endl;
#endif

  if (0 > recvSocket(fd, msgFromService)) {
    cerr << "SrvDrv::localSyncThrSocket "
	 << "read failed from InetSoeketService" << endl;
	  // Mutex̉
#if defined(WIN32)
		ReleaseMutex(that->_mt);
		_endthread();
#else
	  int unlock_status = pthread_mutex_unlock (&(that->_mt));
	  if (0 != unlock_status) {
	    cerr << "[[Error!!]] void* SrvDrv::localSyncThrSocket(void*)" 
			<< endl;
	    cerr << "            pthread_mutex_unlock() failed." << endl;
	    cerr.flush();
		pthread_exit(NULL);
	    return NULL;
	  }
	pthread_exit(NULL);
    return NULL;
#endif
  }

#ifdef DEBUG
  cerr << "** SrvDrv ** READ From InetSocketSrv=[" 
       << msgFromService << "]" << endl;
#endif

  if (msgFromService.length() == 0) {
    cerr << "Received String from Service is Empty" << endl;
    msgFromService = makeResponseIsEmptyMessage();
  }

  //BEGIN: added 2002/07/09
  string fromEnc;
  string toEnc;
  if (getEncodingCharset(msgFromService, fromEnc, toEnc)) {
#ifdef DEBUG
    cerr << "getEncodingCharset: fromEnc=[" << fromEnc << "]: toEnc=[" 
	 << toEnc << "]" << endl;
#endif

    replaceXmlEncoding(msgFromService, toEnc);
#ifdef DEBUG
    cerr << "replaceXmlEncoding:[" << msgFromService << "]" << endl;
#endif

    // convert enc : FROM Content-Type: charset To UTF-8
    convertEncoding(msgFromService, fromEnc, toEnc);
#ifdef DEBUG
    cerr << "msgFromService:[" << msgFromService << "]" << endl;
#endif
  }
  //END: added 2002/07/09

  msgFromService = removeHeaderForService(msgFromService);

  // attach msg ID in the case of async response
  std::string msgIDContent = getMsgIDContent(msgFromMsgDrv);
  if (msgIDContent.length() > 0) {
    attachOrUpdateMsgID(msgFromService, msgIDContent);
  }
  
  // attach backward path, 
  // if this server is the final receiver of async fwd message
  bool itIsAsyncFwdFinalReceiver = 
    isItAsyncFwdFinalReceiver(msgFromMsgDrv, that->localURLs);
  if (itIsAsyncFwdFinalReceiver) {
    attachBackwardPathToResponse(msgFromService, msgFromMsgDrv);
    attachTTLToResponse(msgFromService, msgFromMsgDrv);
  }
  
  that->fileIDOfMsgFromService 
    = convertToFileID(msgFromService, that->srvConf.getSoapMessageSpoolPath());
  
  // Mutex̉
#if defined(WIN32)
	ReleaseMutex(that->_mt);
#else
  int unlock_status = pthread_mutex_unlock (&(that->_mt));
  if (0 != unlock_status) {
    cerr << "[[Error!!]] void* SrvDrv::localSyncThrSocket(void*)" << endl;
    cerr << "            pthread_mutex_unlock() failed." << endl;
    cerr.flush();
	pthread_exit(NULL);
    return NULL;
  }
#endif

#ifdef DEBUG
  cerr << "SrvDrv::localSyncThrSocket: close socket" << endl;
#endif

#if defined(WIN32)
	closesocket(fd);
#else
  close(fd);
#endif

#if defined(WIN32)
	_endthread();
#else
  pthread_exit(NULL);
  return NULL;
#endif
}

//----- [JœIy[Vs߂̃Xbh  -----//
//----- Stdio -----
#if defined(WIN32)
void
#else
void*
#endif
SrvDrv::localSyncThrStdio (void* arg) 
//SrvDrv::localSyncThrFunc (void* arg) 
{
	//pipe read/write index
	const int R = 0;
	const int W = 1;

  SrvDrv* that = (SrvDrv*)arg;
  

#ifdef DEBUG
  // for debug  
  cerr << "SrvDrv::localSyncThrStdio: thr_func wait..." << endl;
#endif

  std::string msgFromService;



  // T[rX̕ԓ҂
	string progname = that->ssmlInfo.getExecProg();

  //std::string hostname = that->ssmlInfo.getHostname();
  //unsigned int port = that->ssmlInfo.getPort();
  
#ifdef DEBUG
  std::cerr << "SrvDrv::localSyncThrStdio: "
	    << "execProg = " << progname << endl;
#endif

//{
#if defined(WIN32)
	// Wo͂ꎞIɃ_CNg
/**
	SrvDrv <==============>    Service
	 pipeout[0]  -----> pipein[0]  = stdin
	 pipein[1]  <-----  pipeout[1] = stdout
*/
	HANDLE pipein[2],pipeout[2];
	SECURITY_ATTRIBUTES sa;
	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	sa.nLength=sizeof(sa);
	sa.lpSecurityDescriptor=NULL;
	sa.bInheritHandle=TRUE;



	if (!CreatePipe(&pipein[0],&pipeout[0],&sa,0)) {



#ifdef DEBUG
		cerr << "SrvDrv::localSyncThrStdio: "
			<< "create pipe0 failed." << endl;
#endif
		msgFromService 
	      = createSOAPFaultMessage("Service.ConnectionPIPENotCreated",
			       "connection pipe not created",
			       "server",
			       that->ssmlInfo.getOperationName() 
			       + " not respond");
	    that->fileIDOfMsgFromService 
	      = convertToFileID(msgFromService, 
				that->srvConf.getSoapMessageSpoolPath());
	    
#if defined(WIN32)
		_endthread();
#else
		pthread_exit(NULL);
	    return NULL;
#endif
	}



	if (!CreatePipe(&pipein[1],&pipeout[1],&sa,0)) {



#ifdef DEBUG
		cerr << "SrvDrv::localSyncThrStdio: "
			<< "create pipe1 failed." << endl;
#endif
		msgFromService 
	      = createSOAPFaultMessage("Service.ConnectionPIPENotCreated",
			       "connection pipe not created",
			       "server",
			       that->ssmlInfo.getOperationName() 
			       + " not respond");
	    that->fileIDOfMsgFromService 
	      = convertToFileID(msgFromService, 
				that->srvConf.getSoapMessageSpoolPath());
#if defined(WIN32)
		CloseHandle(pipein[0]);
		CloseHandle(pipeout[0]);
		_endthread();
#else
		pthread_exit(NULL);
	    return NULL;
#endif
	}


	memset(&si,0,sizeof(si));
	si.cb=sizeof(si);
	si.dwFlags=STARTF_USESTDHANDLES;
	si.hStdInput=pipein[0];
	si.hStdOutput=pipeout[1];
	//si.hStdOutput=GetStdHandle(STD_ERROR_HANDLE);



	if (!CreateProcess(NULL,
					(char*)progname.c_str(),
					NULL,NULL,TRUE,0,NULL,NULL,&si,&pi)) {

		int err=GetLastError();
		if (err==2) {
			cerr << "SrvDrv::localSyncThrStdio: "
				<< "exec prog:" << progname << " not found." << endl;
		}
		else {
			cerr << "SrvDrv::localSyncThrStdio: "
				<< "exec prog:" << progname << " can't exec." << endl;
		}
		msgFromService 
	      = createSOAPFaultMessage("Service.ServiceProcessNotExecuted",
					"service process not executed",
			       "server",
			       that->ssmlInfo.getOperationName() 
			       + " not respond");
	    that->fileIDOfMsgFromService 
	      = convertToFileID(msgFromService, 
				that->srvConf.getSoapMessageSpoolPath());
#if defined(WIN32)
		CloseHandle(pipein[0]);
		CloseHandle(pipeout[1]);
		CloseHandle(pipein[1]);
		CloseHandle(pipeout[0]);

		_endthread();
#else
		pthread_exit(NULL);
	    return NULL;
#endif
	}

	CloseHandle(pipein[0]);
	CloseHandle(pipeout[1]);



#else
	// unix not implemented...
	int pipe_c2p[2];
	int pipe_p2c[2];
	int pid = -1;

	if (pipe(pipe_c2p) < 0) {
		cerr << "SrvDrv::localSyncThrStdio: "
			<< "create pipe 1 failed." << endl;
		pthread_exit(NULL);
	    return NULL;
	}
	if (pipe(pipe_p2c) < 0) {
		cerr << "SrvDrv::localSyncThrStdio: "
			<< "create pipe 2 failed." << endl;
		close(pipe_c2p[R]);
		close(pipe_c2p[W]);
		pthread_exit(NULL);
	    return NULL;
	}

	//qvZX̐
	if ((pid = fork()) < 0) {
		cerr << "SrvDrv::localSyncThrStdio: "
			<< "can't fork child process." << endl;
		close(pipe_c2p[R]);
		close(pipe_c2p[W]);
		close(pipe_p2c[R]);
		close(pipe_p2c[W]);
		pthread_exit(NULL);
	    return NULL;
	}
	//qvZX
	if (0 == pid) {
		close(pipe_p2c[W]);
		close(pipe_c2p[R]);
		dup2(pipe_p2c[R], 0);
		dup2(pipe_c2p[W], 1);
		close(pipe_p2c[R]);
		close(pipe_c2p[W]);
		if (execlp("sh", "sh", "-c", progname.c_str(), NULL) < 0) {
			cerr << "SrvDrv::localSyncThrStdio: "
				<< "can't exec prog : " << progname << endl;
			close(pipe_p2c[R]);
			close(pipe_c2p[W]);
			//pthread_exit(NULL);
		    //return NULL;
			exit(1);
		}
	}

	close(pipe_p2c[R]);
	close(pipe_c2p[W]);
	int wfd = pipe_p2c[W];
	int rfd = pipe_c2p[R];
#endif

//}



	// LOCK
#if defined(WIN32)
	WaitForSingleObject(that->_mt, INFINITE);
#else
int lock_status = pthread_mutex_lock (&(that->_mt));
  if (0 != lock_status) {
    cerr << "[[Error!!]] void* SrvDrv::localSyncThrStdio(void*)" << endl;
    cerr << "            pthread_mutex_lock() failed." << endl;
    cerr.flush();
	pthread_exit(NULL);
    return NULL;
  }
#endif



  // revert file contents from file ID
  std::string msgFromMsgDrv 
    = revertFromFileID(that->fileIDOfMsgFromMsgdrv, 
		       that->srvConf.getSoapMessageSpoolPath());

  // add signature if needed
  if (that->srvConf.isAddSignatureTrue()) {
    if (!addSignatureToString(msgFromMsgDrv,
			      that->srvConf.getSecKeyPath() 
			      + SECURITY_DATA_PRIV_KEY)){
      cerr << "[[Error!!]] void* MsgDrv::localSyncThrStdio(void*)" << endl;
      cerr << "            addSignature failed." << endl;
    };
  }

  // attach HTTP header for service
  std::string msgToService = attachHeaderForService(msgFromMsgDrv);

  // send to service
{
	// message send to service.
#if defined(WIN32)
	unsigned long writeLen = -1;
	WriteFile(pipeout[0], msgToService.c_str(), msgToService.length(), 
			&writeLen, NULL);
#else
	int writeLen = OpenSOAP::write(wfd, msgToService);
#endif



	if (writeLen <= 0) {

	    cerr << "SrvDrv::localSyncThrStdio "
		 << "write failed to Service" << endl;
	  // Mutex̉
#if defined(WIN32)
		CloseHandle(pipein[1]);
		CloseHandle(pipeout[0]);

		ReleaseMutex(that->_mt);
		_endthread();
#else
		  int unlock_status = pthread_mutex_unlock (&(that->_mt));
		  if (0 != unlock_status) {
		    cerr << "[[Error!!]] void* SrvDrv::localSyncThrStdio(void*)" 
				<< endl;
			cerr << "            pthread_mutex_unlock() failed." << endl;
		    cerr.flush();
			pthread_exit(NULL);
		    return NULL;
		  }
		pthread_exit(NULL);
	    return NULL;
#endif
	}
}

#ifdef DEBUG
  cerr << "** SrvDrv ** WRITE TO StdioSrv=[" 
       << msgToService << "]" << endl;
#endif

  {
#if 0
    char buf[2048];
    memset(buf, 0x00, sizeof(buf));
    unsigned long readLen = -1;
    ReadFile(pipein[1], buf, sizeof(buf), &readLen, NULL);
#else
    // message receive from service.
    /*
    vector<char>::size_type	bufLength = 2048L;
    vector<char>	buf(bufLength);
    */
#if defined(WIN32)
    //
    //vector<char>::size_type	bufLength = 2048L;
    vector<char>::size_type	bufLength = 1049600L;
    vector<char>	buf(bufLength);
    //
    unsigned long readLen = -1;
    ReadFile(pipein[1], buf.begin(), buf.size(), &readLen, NULL);
#else
    //int readLen = OpenSOAP::read(rfd, msgFromService);
    int readLen = OpenSOAP::read2(rfd, msgFromService);
#endif

#endif

    if (readLen < 0) {
      
      cerr << "SrvDrv::localSyncThrStdio "
	   << "read failed from InetSoeketService" << endl;
      // Mutex̉
#if defined(WIN32)
      CloseHandle(pipein[1]);
      CloseHandle(pipeout[0]);
      
      ReleaseMutex(that->_mt);
      _endthread();
#else
      int unlock_status = pthread_mutex_unlock (&(that->_mt));
      if (0 != unlock_status) {
	cerr << "[[Error!!]] void* SrvDrv::localSyncThrStdio(void*)" 
	     << endl;
	cerr << "            pthread_mutex_unlock() failed." << endl;
	cerr.flush();
	pthread_exit(NULL);
	return NULL;
      }
      pthread_exit(NULL);
      return NULL;
#endif
    }
#if defined(WIN32)
#if 0
    msgFromService = buf;
#else
    vector<char>::size_type readLength = readLen;
    msgFromService = string(buf.begin(), readLength);
#endif
#endif

#if defined(WIN32)

#ifdef DEBUG
    cerr << "SrvDrv::localSyncThrStdio: close pipe" << endl;
#endif

    CloseHandle(pipein[1]);
    CloseHandle(pipeout[0]);

#else
    close(wfd);
    close(rfd);
#endif
}

#ifdef DEBUG_
  cerr << "** SrvDrv ** READ From StdioSrv=[" 
       << msgFromService << "]" << endl;
#endif

  if (msgFromService.length() == 0) {
    cerr << "Received String from Service is Empty" << endl;
    msgFromService = makeResponseIsEmptyMessage();
  }

  //BEGIN: added 2002/07/09
  string fromEnc;
  string toEnc;
  if (getEncodingCharset(msgFromService, fromEnc, toEnc)) {
#ifdef DEBUG
    cerr << "getEncodingCharset: fromEnc=[" << fromEnc << "]: toEnc=[" 
	 << toEnc << "]" << endl;
#endif

    replaceXmlEncoding(msgFromService, toEnc);
#ifdef DEBUG
    cerr << "replaceXmlEncoding:[" << msgFromService << "]" << endl;
#endif

    // convert enc : FROM Content-Type: charset To UTF-8
    convertEncoding(msgFromService, fromEnc, toEnc);
#ifdef DEBUG
    cerr << "msgFromService:[" << msgFromService << "]" << endl;
#endif
  }
  //END: added 2002/07/09

  msgFromService = removeHeaderForService(msgFromService);

  // attach msg ID in the case of async response
  std::string msgIDContent = getMsgIDContent(msgFromMsgDrv);

  if (msgIDContent.length() > 0) {

    attachOrUpdateMsgID(msgFromService, msgIDContent);
  }

  // attach backward path, 
  // if this server is the final receiver of async fwd message
  bool itIsAsyncFwdFinalReceiver = 
    isItAsyncFwdFinalReceiver(msgFromMsgDrv, that->localURLs);

#ifdef DEBUG
  cerr << "SrvDrv::Stdio-Srv: itIsAsyncFwdFinalReceiver=("
       << itIsAsyncFwdFinalReceiver << ")" << endl;
#endif

  if (itIsAsyncFwdFinalReceiver) {
    attachBackwardPathToResponse(msgFromService, msgFromMsgDrv);
    attachTTLToResponse(msgFromService, msgFromMsgDrv);
  }

  that->fileIDOfMsgFromService 
    = convertToFileID(msgFromService, that->srvConf.getSoapMessageSpoolPath());

  // Mutex̉
#if defined(WIN32)
	ReleaseMutex(that->_mt);
#else
  int unlock_status = pthread_mutex_unlock (&(that->_mt));
  if (0 != unlock_status) {
    cerr << "[[Error!!]] void* SrvDrv::localSyncThrStdio(void*)" << endl;
    cerr << "            pthread_mutex_unlock() failed." << endl;
    cerr.flush();
	pthread_exit(NULL);
    return NULL;
  }
#endif



#if defined(WIN32)
	_endthread();
#else
  pthread_exit(NULL);
  return NULL;
#endif
}

void
SrvDrv::run()
{
#ifdef DEBUG
  cerr << "SVR: read ready." << endl;
#endif

  if (0 > chnlDesc_.read(fileIDOfMsgFromMsgdrv)) {
    cerr << TAG_ERR << "SrvDrv::run: read failed" << endl;
    return;
  }

#ifdef DEBUG
  cerr << "$$ SrvDrv $$ READ From MsgDrv=[" 
       << fileIDOfMsgFromMsgdrv << "]" << endl;
#endif

  soapMsgFromMsgdrv 
    = revertFromFileID(fileIDOfMsgFromMsgdrv, 
		       srvConf.getSoapMessageSpoolPath());

#ifdef DEBUG
  cerr << "extract Operation and Parameter" << endl;
#endif

  std::vector<std::string> operation_names;
  std::vector<std::string> param_names;
  std::vector<std::string> param_values;
  
  string nsStr;
  
  extractOperationAndParam (soapMsgFromMsgdrv, operation_names, 
                            param_names, param_values, &nsStr);
  //                            param_names, param_values);
  
  ssmlInfo.setOperationName(operation_names[0]);
  ssmlInfo.setNamespace(nsStr);
  bool operationExist = getSSMLInfo(ssmlInfo);



  bool soapMsgIsValid = false;
  bool ttlHeaderExist = false;
  std::string ttlType;
  
  soapMsgIsValid 
    = getTTLInfo(soapMsgFromMsgdrv, ttlHeaderExist, ttl, ttlType);

#ifdef DEBUG
  cerr << "SrvDrv::run:ttlHeaderExist = " << ttlHeaderExist << endl;
  cerr << "SrvDrv::run:ttl = " << ttl << endl;
#endif


  if (!soapMsgIsValid) {

    std::string invalidReqErrMsg = makeInvalidRequestMessage();
    std::string fileIDOfInvalidReqErrMsg 
      = convertToFileID(invalidReqErrMsg, srvConf.getSoapMessageSpoolPath());
    chnlDesc_.write(fileIDOfInvalidReqErrMsg); 
    return;
  }
  
  if (!ttlHeaderExist) {
    if (_isAsync) {
      ttl = ssmlInfo.getAsyncTTL();
    }
    else {
      ttl = ssmlInfo.getSyncTTL();
    }
  }
  
  bool targetThreadIsCanceled = false;
   
  std::vector<std::string> soapServerEndPoints;
  getBackwardURLs(soapServerEndPoints);
  
  _itIsAsyncFwdFinalReceiver 
    = isItAsyncFwdFinalReceiver(soapMsgFromMsgdrv, soapServerEndPoints);
  
#ifdef DEBUG
  cerr << "SrvDrv::run: itIsAsyncFwdFinalReceiver=("
       << _itIsAsyncFwdFinalReceiver << ")" << endl;
#endif

  if (operationExist) {

	//################################################
    targetThreadIsCanceled = execLocalSyncOperation();
	//################################################
  }

  if (targetThreadIsCanceled) {

    // ^CAEg thr_func LZďIꍇB
    soapMsgFromService = makeTimeoutMessage();
    fileIDOfMsgFromService 
      = convertToFileID(soapMsgFromService, srvConf.getSoapMessageSpoolPath());
    if (0 > chnlDesc_.write(fileIDOfMsgFromService)) {

      cerr << TAG_ERR << "SrvDrv::run: write failed" << endl;
      return;
    }
  }

  if (_isAsync) {
    if (_itIsAsyncFwdFinalReceiver) {
      // 񓯊]̏ꍇ̓tH[hL[ɑB
#ifdef DEBUG
      cerr << "Response of Async Fwd is send to FwdQueue." << endl;
#endif
      asyncFwdResponse();
    }
    else {
      // 񓯊̏ꍇ̓X|XXv[ɑB
#ifdef DEBUG
      cerr << "Response of Async is send to ResSpool." << endl;
#endif
      asyncResponse();
    }
  }
  else {
    // ̏ꍇ̓bZ[WhCoɃX|XԂB
    if (0 > chnlDesc_.write(fileIDOfMsgFromService)) {

      cerr << TAG_ERR << "SrvDrv::run: write failed" << endl;
      return;
    }
  }
}


void 
SrvDrv::setAsyncFlag(bool isAsync)
{
  _isAsync = isAsync;
}

bool
SrvDrv::asyncResponse()
{
  // Response Spool Ɍʂ𑗂B
  
  ChannelDescriptor resSpoolPushChnlDesc;
  ChannelSelector* resSpoolPushChnlSelector;

#if 0
  std::vector<std::string> backwardURLs;
  getBackwardURLs(backwardURLs);
  if (isItAsyncFwdFinalReceiver(soapMsgFromMsgdrv, backwardURLs)) {
    resSpoolPushChnlSelector = new FwdQueuePushChnlSelector();
  }
  else {
    resSpoolPushChnlSelector = new ResSpoolPushChnlSelector();
  }
#else
    resSpoolPushChnlSelector = new ResSpoolPushChnlSelector();
#endif
  if (0 != resSpoolPushChnlSelector->open(resSpoolPushChnlDesc)) {
    cerr << "SrvDrv::asyncResponse: Chnl open failed" << endl;
    return false;
  }

#ifdef DEBUG
  cerr << "send to response spool" << endl;
#endif

  if (0 > resSpoolPushChnlDesc.write(fileIDOfMsgFromService)) {
    cerr << "SrvDrv::asyncResponse: write failed to Service" << endl;
    return false;
  }
  
  std::string fileIDOfMsgFromResSMgr;
  if (0 > resSpoolPushChnlDesc.read(fileIDOfMsgFromResSMgr)) {
    cerr << "SrvDrv::asyncResponse: read failed from Service" << endl;
    return false;
  }

//#ifdef DEBUG
  cerr << "** SrvDrv ** READ From Service=[" 
       << fileIDOfMsgFromResSMgr << "]" << endl;
//#endif

#ifdef DEBUG
  cerr << "send to request queue" << endl;
#endif

  resSpoolPushChnlSelector->close(resSpoolPushChnlDesc);
  delete resSpoolPushChnlSelector;
  
  std::string msgToReqQMgr("success");
  //std::string fileIDOfMsgToReqQMgr(convertToFileID(msgToReqQMgr));
  //std::string fileIDOfMsgToReqQMgr = convertToFileID(msgToReqQMgr);
  // invalid memory access: string deleted
  std::string fileIDOfMsgToReqQMgr("success");
  if (0 > chnlDesc_.write(fileIDOfMsgToReqQMgr)) {
    cerr << TAG_ERR << "SrvDrv:asyncFwdResponse: write failed" << endl;
    return false;
  }
  return true;
}

//----- 񓯊]̍ŏIT[o[̏ꍇ̃X|Xs -----//
bool
SrvDrv::asyncFwdResponse()
{
  // 񓯊]̍Ō̃T[o[̏ꍇ
  // Fwd Queue Ɍʂ𑗂B
  
  ChannelDescriptor fwdQueuePushChnlDesc;
  ChannelSelector* fwdQueuePushChnlSelector = new FwdQueuePushChnlSelector();
  if (0 != fwdQueuePushChnlSelector->open(fwdQueuePushChnlDesc)) {
    cerr << "SrvDrv::asyncFwdResponse:: Chnl open failed" << endl;
    return false;
  }

#ifdef DEBUG
  cerr << "send to response spool" << endl;
#endif

  if (0 > fwdQueuePushChnlDesc.write(fileIDOfMsgFromService)) {
    cerr << "SrvDrv::asyncFwdResponse: write failed to Service" << endl;
    return false;
  }
  
  std::string fileIDOfMsgFromResSMgr;
  if (0 > fwdQueuePushChnlDesc.read(fileIDOfMsgFromResSMgr)) {
    cerr << "SrvDrv::asyncFwdResponse: read failed from Service" << endl;
    return false;
  }

//#ifdef DEBUG
  cerr << "** SrvDrv ** READ From Service=[" 
       << fileIDOfMsgFromResSMgr << "]" << endl;
  cerr << "send to request queue" << endl;
//#endif

  fwdQueuePushChnlSelector->close(fwdQueuePushChnlDesc);
  delete fwdQueuePushChnlSelector;
  
  
  std::string msgToReqQMgr("success");
  //std::string fileIDOfMsgToReqQMgr = convertToFileID(msgToReqQMgr);
  std::string fileIDOfMsgToReqQMgr("success");
  if (0 > chnlDesc_.write(fileIDOfMsgToReqQMgr)) {
    cerr << TAG_ERR << "SrvDrv::asyncFwdResponse: write failed" << endl;
    return false;
  }
  
  return true;
}
